"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Union

from ..models import ResourceGroup, ResourceType
from ..models.light.configuration_tool import ConfigurationTool
from ..models.light.cover import Cover, CoverAction
from ..models.light.light import Light, LightAlert, LightEffect, LightFanSpeed
from ..models.light.lock import Lock
from ..models.light.range_extender import RangeExtender
from ..models.light.siren import Siren
from .api_handlers import APIHandler, GroupedAPIHandler

if TYPE_CHECKING:
    from ..gateway import DeconzSession

LOGGER = logging.getLogger(__name__)


class ConfigurationToolHandler(APIHandler[ConfigurationTool]):
    """Handler for configuration tool."""

    resource_group = ResourceGroup.LIGHT
    resource_type = ResourceType.CONFIGURATION_TOOL
    item_cls = ConfigurationTool


class CoverHandler(APIHandler[Cover]):
    """Handler for covers."""

    resource_group = ResourceGroup.LIGHT
    resource_types = {
        ResourceType.LEVEL_CONTROLLABLE_OUTPUT,
        ResourceType.WINDOW_COVERING_CONTROLLER,
        ResourceType.WINDOW_COVERING_DEVICE,
    }
    item_cls = Cover

    async def set_state(
        self,
        id: str,
        action: CoverAction | None = None,
        lift: int | None = None,
        tilt: int | None = None,
        legacy_mode: bool = False,
    ) -> dict[str, Any]:
        """Set state of cover.

        Supported values:
        - action [CoverAction] Open, Close, Stop
        - lift [int] between 0-100
        - tilt [int] between 0-100
        """
        data: dict[str, bool | int] = {}

        if action is not None and not legacy_mode:
            if action is CoverAction.OPEN:
                data["open"] = True
            elif action is CoverAction.CLOSE:
                data["open"] = False
            elif action is CoverAction.STOP:
                data["stop"] = True
        elif action is not None and legacy_mode:
            if action is CoverAction.OPEN:
                data["on"] = False
            elif action is CoverAction.CLOSE:
                data["on"] = True
            elif action is CoverAction.STOP:
                data["bri_inc"] = 0
        elif not legacy_mode:
            if lift is not None:
                data["lift"] = lift
            if tilt is not None:
                data["tilt"] = tilt
        else:
            if lift is not None:
                data["bri"] = int(lift * 2.54)
            if tilt is not None:
                data["sat"] = int(tilt * 2.54)

        return await self.gateway.request_with_retry(
            "put",
            path=f"{self.path}/{id}/state",
            json=data,
        )


class LightHandler(APIHandler[Light]):
    """Handler for lights."""

    resource_group = ResourceGroup.LIGHT
    resource_types = {
        ResourceType.COLOR_DIMMABLE_LIGHT,
        ResourceType.COLOR_LIGHT,
        ResourceType.COLOR_TEMPERATURE_LIGHT,
        ResourceType.EXTENDED_COLOR_LIGHT,
        ResourceType.DIMMABLE_LIGHT,
        ResourceType.DIMMABLE_PLUGIN_UNIT,
        ResourceType.DIMMER_SWITCH,
        ResourceType.FAN,
        ResourceType.ON_OFF_LIGHT,
        ResourceType.ON_OFF_OUTPUT,
        ResourceType.ON_OFF_PLUGIN_UNIT,
        ResourceType.SMART_PLUG,
        ResourceType.UNKNOWN,  # Legacy support
    }
    item_cls = Light

    async def set_state(
        self,
        id: str,
        alert: LightAlert | None = None,
        brightness: int | None = None,
        color_loop_speed: int | None = None,
        color_temperature: int | None = None,
        effect: LightEffect | None = None,
        fan_speed: LightFanSpeed | None = None,
        hue: int | None = None,
        on: bool | None = None,
        on_time: int | None = None,
        saturation: int | None = None,
        transition_time: int | None = None,
        xy: tuple[float, float] | None = None,
    ) -> dict[str, Any]:
        """Change state of a light.

        Supported values:
        - alert [str]
          - "none" light is not performing an alert
          - "select" light is blinking a short time
          - "lselect" light is blinking a longer time
        - brightness [int] 0-255
        - color_loop_speed [int] 1-255
          - 1 = very fast
          - 15 is default
          - 255 very slow
        - color_temperature [int] between ctmin-ctmax
        - effect [str]
          - "none" no effect
          - "colorloop" the light will cycle continuously through all
                        colors with the speed specified by colorloopspeed
        - fan_speed [FanSpeed] Off, 25%, 50%, 75%, 100%, Auto, ComfortBreeze
        - hue [int] 0-65535
        - on [bool] True/False
        - on_time [int] 0-65535 1/10 seconds resolution
        - saturation [int] 0-255
        - transition_time [int] 0-65535 1/10 seconds resolution
        - xy [tuple] (0-1, 0-1)
        """
        data: dict[str, Any] = {
            key: value
            for key, value in {
                "bri": brightness,
                "colorloopspeed": color_loop_speed,
                "ct": color_temperature,
                "hue": hue,
                "on": on,
                "ontime": on_time,
                "sat": saturation,
                "transitiontime": transition_time,
                "xy": xy,
            }.items()
            if value is not None
        }
        if alert is not None:
            data["alert"] = alert.value
        if effect is not None:
            data["effect"] = effect.value
        if fan_speed is not None:
            data["speed"] = fan_speed.value
        return await self.gateway.request_with_retry(
            "put",
            path=f"{self.path}/{id}/state",
            json=data,
        )


class LockHandler(APIHandler[Lock]):
    """Handler for fans."""

    resource_group = ResourceGroup.LIGHT
    resource_type = ResourceType.DOOR_LOCK
    item_cls = Lock

    async def set_state(self, id: str, lock: bool) -> dict[str, Any]:
        """Set state of lock.

        Supported values:
        - lock [bool] True/False.
        """
        return await self.gateway.request_with_retry(
            "put",
            path=f"{self.path}/{id}/state",
            json={"on": lock},
        )


class RangeExtenderHandler(APIHandler[ConfigurationTool]):
    """Handler for range extender."""

    resource_group = ResourceGroup.LIGHT
    resource_type = ResourceType.RANGE_EXTENDER
    item_cls = RangeExtender


class SirenHandler(APIHandler[Siren]):
    """Handler for sirens."""

    resource_group = ResourceGroup.LIGHT
    resource_type = ResourceType.WARNING_DEVICE
    item_cls = Siren

    async def set_state(
        self,
        id: str,
        on: bool,
        duration: int | None = None,
    ) -> dict[str, Any]:
        """Turn on device.

        Supported values:
        - on [bool] True/False
        - duration [int] 1/10th of a second
        """
        data: dict[str, int | str] = {}

        data["alert"] = (LightAlert.LONG if on else LightAlert.NONE).value
        if on and duration is not None:
            data["ontime"] = duration

        return await self.gateway.request_with_retry(
            "put",
            path=f"{self.path}/{id}/state",
            json=data,
        )


LightResources = Union[
    ConfigurationTool,
    Cover,
    Light,
    Lock,
    Siren,
]


class LightResourceManager(GroupedAPIHandler[LightResources]):
    """Represent deCONZ lights."""

    resource_group = ResourceGroup.LIGHT

    def __init__(self, gateway: DeconzSession) -> None:
        """Initialize light manager."""
        self.configuration_tool = ConfigurationToolHandler(gateway, grouped=True)
        self.covers = CoverHandler(gateway, grouped=True)
        self.lights = LightHandler(gateway, grouped=True)
        self.locks = LockHandler(gateway, grouped=True)
        self.range_extender = RangeExtenderHandler(gateway, grouped=True)
        self.sirens = SirenHandler(gateway, grouped=True)

        handlers: list[APIHandler[Any]] = [
            self.configuration_tool,
            self.covers,
            self.lights,
            self.locks,
            self.range_extender,
            self.sirens,
        ]

        super().__init__(gateway, handlers)



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

from typing import TypedDict

from . import SensorBase


class TypedMoistureState(TypedDict):
    """Moisture state type definition."""

    moisture: int


class TypedMoisture(TypedDict):
    """Moisture type definition."""

    state: TypedMoistureState


class Moisture(SensorBase):
    """Moisture sensor."""

    raw: TypedMoisture

    @property
    def moisture(self) -> int:
        """Moisture level.

        0-100 in percent.
        """
        return self.raw["state"]["moisture"]



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

from typing import Any

from ..models import ResourceGroup, ResourceType
from ..models.group import Group
from ..models.light.light import LightAlert, LightEffect
from .api_handlers import APIHandler


class GroupHandler(APIHandler[Group]):
    """Represent deCONZ groups."""

    resource_group = ResourceGroup.GROUP
    resource_type = ResourceType.GROUP
    item_cls = Group

    async def set_attributes(
        self,
        id: str,
        hidden: bool | None = None,
        light_sequence: list[str] | None = None,
        lights: list[str] | None = None,
        multi_device_ids: list[str] | None = None,
        name: str | None = None,
    ) -> dict[str, Any]:
        """Change attributes of a group.

        Supported values:
        - hidden [bool] Indicates the hidden status of the group
        - light_sequence [list of light IDs] Specify a sorted list of light IDs for apps
        - lights [list of light IDs]IDs of the lights which are members of the group
        - multi_device_ids [int] Subsequential light IDs of multidevices
        - name [str] The name of the group
        """
        data = {
            key: value
            for key, value in {
                "hidden": hidden,
                "lightsequence": light_sequence,
                "lights": lights,
                "multideviceids": multi_device_ids,
                "name": name,
            }.items()
            if value is not None
        }
        return await self.gateway.request(
            "put",
            path=f"{self.path}/{id}",
            json=data,
        )

    async def set_state(
        self,
        id: str,
        alert: LightAlert | None = None,
        brightness: int | None = None,
        color_loop_speed: int | None = None,
        color_temperature: int | None = None,
        effect: LightEffect | None = None,
        hue: int | None = None,
        on: bool | None = None,
        on_time: int | None = None,
        saturation: int | None = None,
        toggle: bool | None = None,
        transition_time: int | None = None,
        xy: tuple[float, float] | None = None,
    ) -> dict[str, Any]:
        """Change state of a group.

        Supported values:
        - alert [str]
          - "none" light is not performing an alert
          - "select" light is blinking a short time
          - "lselect" light is blinking a longer time
        - brightness [int] 0-255
        - color_loop_speed [int] 1-255
          - 1 = very fast
          - 15 is default
          - 255 very slow
        - color_temperature [int] between ctmin-ctmax
        - effect [str]
          - "none" no effect
          - "colorloop" the light will cycle continuously through all
                        colors with the speed specified by colorloopspeed
        - hue [int] 0-65535
        - on [bool] True/False
        - on_time [int] 0-65535 1/10 seconds resolution
        - saturation [int] 0-255
        - toggle [bool] True toggles the lights of that group from on to off
                        or vice versa, false has no effect
        - transition_time [int] 0-65535 1/10 seconds resolution
        - xy [tuple] (0-1, 0-1)
        """
        data: dict[str, Any] = {
            key: value
            for key, value in {
                "bri": brightness,
                "colorloopspeed": color_loop_speed,
                "ct": color_temperature,
                "hue": hue,
                "on": on,
                "ontime": on_time,
                "sat": saturation,
                "toggle": toggle,
                "transitiontime": transition_time,
                "xy": xy,
            }.items()
            if value is not None
        }
        if alert is not None:
            data["alert"] = alert.value
        if effect is not None:
            data["effect"] = effect.value
        return await self.gateway.request_with_retry(
            "put",
            path=f"{self.path}/{id}/action",
            json=data,
        )



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

import enum
import logging
from typing import Literal, TypedDict

from . import SensorBase

LOGGER = logging.getLogger(__name__)


class AncillaryControlAction(enum.Enum):
    """Last action a user invoked on the keypad."""

    ARMED_AWAY = "armed_away"
    ARMED_NIGHT = "armed_night"
    ARMED_STAY = "armed_stay"
    DISARMED = "disarmed"
    EMERGENCY = "emergency"
    FIRE = "fire"
    INVALID_CODE = "invalid_code"
    PANIC = "panic"


class AncillaryControlPanel(enum.Enum):
    """Mirror of alarm system state.armstate attribute."""

    ARMED_AWAY = "armed_away"
    ARMED_NIGHT = "armed_night"
    ARMED_STAY = "armed_stay"
    ARMING_AWAY = "arming_away"
    ARMING_NIGHT = "arming_night"
    ARMING_STAY = "arming_stay"
    DISARMED = "disarmed"
    ENTRY_DELAY = "entry_delay"
    EXIT_DELAY = "exit_delay"
    IN_ALARM = "in_alarm"
    NOT_READY = "not_ready"

    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value: object) -> "AncillaryControlPanel":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unexpected panel mode %s", value)
        return AncillaryControlPanel.UNKNOWN


class TypedAncillaryControlState(TypedDict):
    """Ancillary control state type definition."""

    action: Literal[
        "armed_away",
        "armed_night",
        "armed_stay",
        "disarmed",
        "emergency",
        "fire",
        "invalid_code",
        "panic",
    ]
    panel: Literal[
        "armed_away",
        "armed_night",
        "armed_stay",
        "arming_away",
        "arming_night",
        "arming_stay",
        "disarmed",
        "entry_delay",
        "exit_delay",
        "in_alarm",
        "not_ready",
    ]
    seconds_remaining: int


class TypedAncillaryControl(TypedDict):
    """Ancillary control type definition."""

    state: TypedAncillaryControlState


class AncillaryControl(SensorBase):
    """Ancillary control sensor."""

    raw: TypedAncillaryControl

    @property
    def action(self) -> AncillaryControlAction:
        """Last action a user invoked on the keypad."""
        return AncillaryControlAction(self.raw["state"]["action"])

    @property
    def panel(self) -> AncillaryControlPanel | None:
        """Mirror of alarm system state.armstate attribute.

        It reflects what is shown on the panel (when activated by the keypad’s proximity sensor).
        """
        if "panel" in self.raw["state"]:
            return AncillaryControlPanel(self.raw["state"]["panel"])
        return None

    @property
    def seconds_remaining(self) -> int:
        """Remaining time of "exit_delay" and "entry_delay" states.

        In all other states the value is 0.
        """
        return self.raw["state"].get("seconds_remaining", 0)



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

import enum
import logging
from typing import Literal, TypedDict

from . import SensorBase

LOGGER = logging.getLogger(__name__)


class ThermostatFanMode(enum.Enum):
    """Fan mode.

    Supported values:
    - "off"
    - "low"
    - "medium"
    - "high"
    - "on"
    - "auto"
    - "smart"
    """

    OFF = "off"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    ON = "on"
    AUTO = "auto"
    SMART = "smart"

    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value: object) -> "ThermostatFanMode":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unexpected thermostat fan mode %s", value)
        return ThermostatFanMode.UNKNOWN


class ThermostatMode(enum.Enum):
    """Set the current operating mode of a thermostat.

    Supported values:
    - "off"
    - "auto"
    - "cool"
    - "heat"
    - "emergency heating"
    - "precooling"
    - "fan only"
    - "dry"
    - "sleep"
    """

    OFF = "off"
    AUTO = "auto"
    COOL = "cool"
    HEAT = "heat"
    EMERGENCY_HEATING = "emergency heating"
    PRE_COOLING = "precooling"
    FAN_ONLY = "fan only"
    DRY = "dry"
    SLEEP = "sleep"

    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value: object) -> "ThermostatMode":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unexpected thermostat mode %s", value)
        return ThermostatMode.UNKNOWN


class ThermostatSwingMode(enum.Enum):
    """Set the AC louvers position.

    Supported values:
    - "fully closed"
    - "fully open"
    - "quarter open"
    - "half open"
    - "three quarters open"
    """

    FULLY_CLOSED = "fully closed"
    FULLY_OPEN = "fully open"
    QUARTER_OPEN = "quarter open"
    HALF_OPEN = "half open"
    THREE_QUARTERS_OPEN = "three quarters open"

    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value: object) -> "ThermostatSwingMode":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unexpected thermostat swing mode %s", value)
        return ThermostatSwingMode.UNKNOWN


class ThermostatPreset(enum.Enum):
    """Set the current operating mode for Tuya thermostats.

    Supported values:
    - "auto"
    - "boost"
    - "comfort"
    - "complex"
    - "eco"
    - "holiday"
    - "manual"
    - "program"
    """

    AUTO = "auto"
    BOOST = "boost"
    COMFORT = "comfort"
    COMPLEX = "complex"
    ECO = "eco"
    HOLIDAY = "holiday"
    MANUAL = "manual"
    PROGRAM = "program"

    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value: object) -> "ThermostatPreset":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unexpected thermostat preset %s", value)
        return ThermostatPreset.UNKNOWN


class ThermostatTemperatureMeasurement(enum.Enum):
    """Set the mode of operation for Elko Super TR thermostat.

    Supported values:
    - "air sensor"
    - "floor sensor"
    - "floor protection"
    """

    AIR_SENSOR = "air sensor"
    FLOOR_SENSOR = "floor sensor"
    FLOOR_PROTECTION = "floor protection"

    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value: object) -> "ThermostatTemperatureMeasurement":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unexpected thermostat temperature measurement %s", value)
        return ThermostatTemperatureMeasurement.UNKNOWN


class TypedThermostatConfig(TypedDict):
    """Thermostat config type definition."""

    coolsetpoint: int
    displayflipped: bool
    externalsensortemp: int
    externalwindowopen: bool
    fanmode: Literal["off", "low", "medium", "high", "on", "auto", "smart"]
    heatsetpoint: int
    locked: bool
    mode: Literal[
        "off",
        "auto",
        "cool",
        "heat",
        "emergency heating",
        "precooling",
        "fan only",
        "dry",
        "sleep",
    ]
    mountingmode: bool
    offset: int
    preset: Literal[
        "auto", "boost", "comfort", "complex", "eco", "holiday", "manual", "program"
    ]
    schedule_on: bool
    swingmode: Literal[
        "fully closed", "fully open", "quarter open", "half open", "three quarters open"
    ]
    temperaturemeasurement: Literal["air sensor", "floor sensor", "floor protection"]
    windowopen_set: bool


class TypedThermostatState(TypedDict):
    """Thermostat state type definition."""

    errorcode: bool
    floortemperature: int
    heating: bool
    mountingmodeactive: bool
    on: bool
    temperature: int
    valve: int


class TypedThermostat(TypedDict):
    """Thermostat type definition."""

    config: TypedThermostatConfig
    state: TypedThermostatState


class Thermostat(SensorBase):
    """Thermostat "sensor"."""

    raw: TypedThermostat

    @property
    def cooling_setpoint(self) -> int | None:
        """Cooling setpoint.

        700-3500.
        """
        return self.raw["config"].get("coolsetpoint")

    @property
    def scaled_cooling_setpoint(self) -> float | None:
        """Cooling setpoint.

        7-35.
        """
        if temperature := self.cooling_setpoint:
            return round(temperature / 100, 1)
        return None

    @property
    def display_flipped(self) -> bool | None:
        """Tells if display for TRVs is flipped."""
        return self.raw["config"].get("displayflipped")

    @property
    def error_code(self) -> bool | None:
        """Error code."""
        return self.raw["state"].get("errorcode")

    @property
    def external_sensor_temperature(self) -> int | None:
        """Track temperature value provided by an external sensor.

        -32768–32767.
        Device dependent and only exposed for devices supporting it.
        """
        return self.raw["config"].get("externalsensortemp")

    @property
    def scaled_external_sensor_temperature(self) -> float | None:
        """Track temperature value provided by an external sensor.

        -327–327.
        """
        if temperature := self.external_sensor_temperature:
            return round(temperature / 100, 1)
        return None

    @property
    def external_window_open(self) -> bool | None:
        """Track open/close state of an external sensor.

        Device dependent and only exposed for devices supporting it.
        """
        return self.raw["config"].get("externalwindowopen")

    @property
    def fan_mode(self) -> ThermostatFanMode | None:
        """Fan mode."""
        if "fanmode" in self.raw["config"]:
            return ThermostatFanMode(self.raw["config"]["fanmode"])
        return None

    @property
    def floor_temperature(self) -> int | None:
        """Floor temperature."""
        return self.raw["state"].get("floortemperature")

    @property
    def scaled_floor_temperature(self) -> float | None:
        """Floor temperature."""
        if temperature := self.floor_temperature:
            return round(temperature / 100, 1)
        return None

    @property
    def heating(self) -> bool | None:
        """Heating setpoint."""
        return self.raw["state"].get("heating")

    @property
    def heating_setpoint(self) -> int | None:
        """Heating setpoint.

        500-3200.
        """
        return self.raw["config"].get("heatsetpoint")

    @property
    def scaled_heating_setpoint(self) -> float | None:
        """Heating setpoint.

        5-32.
        """
        if temperature := self.heating_setpoint:
            return round(temperature / 100, 1)
        return None

    @property
    def locked(self) -> bool | None:
        """Child lock active/inactive for thermostats/TRVs supporting it."""
        return self.raw["config"].get("locked")

    @property
    def mode(self) -> ThermostatMode | None:
        """Set the current operating mode of a thermostat."""
        if "mode" in self.raw["config"]:
            return ThermostatMode(self.raw["config"]["mode"])
        return None

    @property
    def mounting_mode(self) -> bool | None:
        """Set a TRV into mounting mode if supported (valve fully open position)."""
        return self.raw["config"].get("mountingmode")

    @property
    def mounting_mode_active(self) -> bool | None:
        """If thermostat mounting mode is active."""
        return self.raw["state"].get("mountingmodeactive")

    @property
    def offset(self) -> int | None:
        """Add a signed offset value to measured temperature and humidity state values.

        Values send by the REST-API are already amended by the offset.
        """
        return self.raw["config"].get("offset")

    @property
    def preset(self) -> ThermostatPreset | None:
        """Set the current operating mode for Tuya thermostats."""
        if "preset" in self.raw["config"]:
            return ThermostatPreset(self.raw["config"]["preset"])
        return None

    @property
    def schedule_enabled(self) -> bool | None:
        """Tell when thermostat schedule is enabled."""
        return self.raw["config"].get("schedule_on")

    @property
    def state_on(self) -> bool | None:
        """Declare if the sensor is on or off."""
        return self.raw["state"].get("on")

    @property
    def swing_mode(self) -> ThermostatSwingMode | None:
        """Set the AC louvers position."""
        if "swingmode" in self.raw["config"]:
            return ThermostatSwingMode(self.raw["config"]["swingmode"])
        return None

    @property
    def temperature(self) -> int:
        """Temperature."""
        return self.raw["state"]["temperature"]

    @property
    def scaled_temperature(self) -> float:
        """Scaled temperature."""
        return round(self.temperature / 100, 1)

    @property
    def temperature_measurement(self) -> ThermostatTemperatureMeasurement | None:
        """Set the mode of operation for Elko Super TR thermostat."""
        if "temperaturemeasurement" in self.raw["config"]:
            return ThermostatTemperatureMeasurement(
                self.raw["config"]["temperaturemeasurement"]
            )
        return None

    @property
    def valve(self) -> int | None:
        """How open the valve is."""
        return self.raw["state"].get("valve")

    @property
    def window_open_detection(self) -> bool | None:
        """Set if window open detection shall be active or inactive for Tuya thermostats.

        Device dependent and only exposed for devices supporting it.
        """
        return self.raw["config"].get("windowopen_set")



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

import enum
from typing import TypedDict

from . import SensorBase


class TypedRelativeRotaryState(TypedDict):
    """Relative rotary state type definition."""

    expectedeventduration: int
    expectedrotation: int
    rotaryevent: int


class TypedRelativeRotary(TypedDict):
    """Relative rotary type definition."""

    state: TypedRelativeRotaryState


class RelativeRotaryEvent(enum.IntEnum):
    """Rotary event.

    Supported values:
    - 1 - new movements (start)
    - 2 - repeat movements
    """

    NEW = 1
    REPEAT = 2


class RelativeRotary(SensorBase):
    """Relative rotary sensor."""

    raw: TypedRelativeRotary

    @property
    def expected_event_duration(self) -> int:
        """Event duration to expect.

        Interval [ms] which rotary events will be emit.
        """
        return self.raw["state"]["expectedeventduration"]

    @property
    def expected_rotation(self) -> int:
        """Rotation to expect.

        Report angle
        - positive for clockwise
        - negative for counter-clockwise
        """
        return self.raw["state"]["expectedrotation"]

    @property
    def rotary_event(self) -> RelativeRotaryEvent:
        """Rotary event.

        - 1 for new movements (start)
        - 2 for repeat movements
        """
        return RelativeRotaryEvent(self.raw["state"]["rotaryevent"])



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

from .. import ResourceGroup
from ..deconz_device import DeconzDevice


class SensorBase(DeconzDevice):
    """deCONZ sensor representation.

    Dresden Elektroniks documentation of sensors in deCONZ
    http://dresden-elektronik.github.io/deconz-rest-doc/sensors/
    """

    resource_group = ResourceGroup.SENSOR

    @property
    def battery(self) -> int | None:
        """Battery status of sensor."""
        raw: dict[str, int] = self.raw["config"]
        return raw.get("battery")

    @property
    def ep(self) -> int | None:
        """Endpoint of sensor."""
        raw: dict[str, int] = self.raw
        return raw.get("ep")

    @property
    def low_battery(self) -> bool | None:
        """Low battery."""
        raw: dict[str, bool] = self.raw["state"]
        return raw.get("lowbattery")

    @property
    def on(self) -> bool | None:
        """Declare if the sensor is on or off."""
        raw: dict[str, bool] = self.raw["config"]
        return raw.get("on")

    @property
    def reachable(self) -> bool:
        """Declare if the sensor is reachable."""
        raw: dict[str, bool] = self.raw["config"]
        return raw.get("reachable", True)

    @property
    def tampered(self) -> bool | None:
        """Tampered."""
        raw: dict[str, bool] = self.raw["state"]
        return raw.get("tampered")

    @property
    def internal_temperature(self) -> float | None:
        """Extra temperature available on some Xiaomi devices."""
        raw: dict[str, int] = self.raw["config"]
        if temperature := raw.get("temperature"):
            return round(temperature / 100, 1)
        return None



"""Python library to connect deCONZ and Home Assistant to work together."""

from typing import TypedDict

from . import SensorBase


class TypedPressureState(TypedDict):
    """Pressure state type definition."""

    pressure: int


class TypedPressure(TypedDict):
    """Pressure type definition."""

    state: TypedPressureState


class Pressure(SensorBase):
    """Pressure sensor."""

    raw: TypedPressure

    @property
    def pressure(self) -> int:
        """Pressure."""
        return self.raw["state"]["pressure"]



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union

from ..models import ResourceGroup, ResourceType
from ..models.sensor.air_purifier import AirPurifier, AirPurifierFanMode
from ..models.sensor.air_quality import AirQuality
from ..models.sensor.alarm import Alarm
from ..models.sensor.ancillary_control import AncillaryControl
from ..models.sensor.battery import Battery
from ..models.sensor.carbon_monoxide import CarbonMonoxide
from ..models.sensor.consumption import Consumption
from ..models.sensor.daylight import Daylight
from ..models.sensor.door_lock import DoorLock
from ..models.sensor.fire import Fire
from ..models.sensor.generic_flag import GenericFlag
from ..models.sensor.generic_status import GenericStatus
from ..models.sensor.humidity import Humidity
from ..models.sensor.light_level import LightLevel
from ..models.sensor.moisture import Moisture
from ..models.sensor.open_close import OpenClose
from ..models.sensor.power import Power
from ..models.sensor.presence import (
    Presence,
    PresenceConfigDeviceMode,
    PresenceConfigTriggerDistance,
)
from ..models.sensor.pressure import Pressure
from ..models.sensor.relative_rotary import RelativeRotary
from ..models.sensor.switch import (
    Switch,
    SwitchDeviceMode,
    SwitchMode,
    SwitchWindowCoveringType,
)
from ..models.sensor.temperature import Temperature
from ..models.sensor.thermostat import (
    Thermostat,
    ThermostatFanMode,
    ThermostatMode,
    ThermostatPreset,
    ThermostatSwingMode,
    ThermostatTemperatureMeasurement,
)
from ..models.sensor.time import Time
from ..models.sensor.vibration import Vibration
from ..models.sensor.water import Water
from .api_handlers import APIHandler, GroupedAPIHandler

if TYPE_CHECKING:
    from ..gateway import DeconzSession


class AirPurifierHandler(APIHandler[AirPurifier]):
    """Handler for air purifier sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_AIR_PURIFIER
    item_cls = AirPurifier

    async def set_config(
        self,
        id: str,
        fan_mode: AirPurifierFanMode | None = None,
        filter_life_time: int | None = None,
        led_indication: bool | None = None,
        locked: bool | None = None,
    ) -> dict[str, Any]:
        """Set speed of fans/ventilators.

        Supported values:
        - fan_mode [AirPurifierFanMode]
          - "off"
          - "auto"
          - "speed_1"
          - "speed_2"
          - "speed_3"
          - "speed_4"
          - "speed_5"
        - filter_life_time [int] 0-65536
        - led_indication [bool] True/False
        - locked [bool] True/False
        """
        data: dict[str, int | str] = {
            key: value
            for key, value in {
                "filterlifetime": filter_life_time,
                "ledindication": led_indication,
                "locked": locked,
            }.items()
            if value is not None
        }
        if fan_mode is not None:
            data["mode"] = fan_mode.value

        return await self.gateway.request_with_retry(
            "put",
            path=f"{self.path}/{id}/config",
            json=data,
        )


class AirQualityHandler(APIHandler[AirQuality]):
    """Handler for air quality sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_AIR_QUALITY
    item_cls = AirQuality


class AlarmHandler(APIHandler[Alarm]):
    """Handler for alarm sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_ALARM
    item_cls = Alarm


class AncillaryControlHandler(APIHandler[AncillaryControl]):
    """Handler for ancillary control sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_ANCILLARY_CONTROL
    item_cls = AncillaryControl


class BatteryHandler(APIHandler[Battery]):
    """Handler for battery sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_BATTERY
    item_cls = Battery


class CarbonMonoxideHandler(APIHandler[CarbonMonoxide]):
    """Handler for carbon monoxide sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_CARBON_MONOXIDE
    item_cls = CarbonMonoxide


class ConsumptionHandler(APIHandler[Consumption]):
    """Handler for consumption sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_CONSUMPTION
    item_cls = Consumption


class DaylightHandler(APIHandler[Daylight]):
    """Handler for daylight sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.DAYLIGHT
    item_cls = Daylight

    async def set_config(
        self,
        id: str,
        sunrise_offset: int | None = None,
        sunset_offset: int | None = None,
    ) -> dict[str, Any]:
        """Set config of the daylight sensor.

        Supported values:
        - sunrise_offset [int] -120-120
        - sunset_offset [int] -120-120
        """
        data: dict[str, int] = {
            key: value
            for key, value in {
                "sunriseoffset": sunrise_offset,
                "sunsetoffset": sunset_offset,
            }.items()
            if value is not None
        }
        return await self.gateway.request_with_retry(
            "put",
            path=f"{self.path}/{id}/config",
            json=data,
        )


class DoorLockHandler(APIHandler[DoorLock]):
    """Handler for door lock sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_DOOR_LOCK
    item_cls = DoorLock

    async def set_config(self, id: str, lock: bool) -> dict[str, Any]:
        """Set config of the lock.

        Supported values:
        - Lock [bool] True/False.
        """
        return await self.gateway.request_with_retry(
            "put",
            path=f"{self.path}/{id}/config",
            json={"lock": lock},
        )


class FireHandler(APIHandler[Fire]):
    """Handler for fire sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_FIRE
    item_cls = Fire


class GenericFlagHandler(APIHandler[GenericFlag]):
    """Handler for generic flag sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.CLIP_GENERIC_FLAG
    item_cls = GenericFlag


class GenericStatusHandler(APIHandler[GenericStatus]):
    """Handler for generic status sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.CLIP_GENERIC_STATUS
    item_cls = GenericStatus


class HumidityHandler(APIHandler[Humidity]):
    """Handler for humidity sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_types = {
        ResourceType.ZHA_HUMIDITY,
        ResourceType.CLIP_HUMIDITY,
    }
    item_cls = Humidity

    async def set_config(self, id: str, offset: int) -> dict[str, Any]:
        """Change config of humidity sensor.

        Supported values:
        - offset [int] -32768–32767
        """
        return await self.gateway.request_with_retry(
            "put",
            path=f"{self.path}/{id}/config",
            json={"offset": offset},
        )


class LightLevelHandler(APIHandler[LightLevel]):
    """Handler for light level sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_types = {
        ResourceType.ZHA_LIGHT_LEVEL,
        ResourceType.CLIP_LIGHT_LEVEL,
    }
    item_cls = LightLevel

    async def set_config(
        self,
        id: str,
        threshold_dark: int | None = None,
        threshold_offset: int | None = None,
    ) -> dict[str, Any]:
        """Change config of presence sensor.

        Supported values:
        - threshold_dark [int] 0-65534
        - threshold_offset [int] 1-65534
        """
        data = {
            key: value
            for key, value in {
                "tholddark": threshold_dark,
                "tholdoffset": threshold_offset,
            }.items()
            if value is not None
        }
        return await self.gateway.request_with_retry(
            "put",
            path=f"{self.path}/{id}/config",
            json=data,
        )


class MoistureHandler(APIHandler[Moisture]):
    """Handler for moisture sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_MOISTURE
    item_cls = Moisture


class OpenCloseHandler(APIHandler[OpenClose]):
    """Handler for open/close sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_types = {
        ResourceType.ZHA_OPEN_CLOSE,
        ResourceType.CLIP_OPEN_CLOSE,
    }
    item_cls = OpenClose


class PowerHandler(APIHandler[Power]):
    """Handler for power sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_POWER
    item_cls = Power


class PresenceHandler(APIHandler[Presence]):
    """Handler for presence sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_types = {
        ResourceType.ZHA_PRESENCE,
        ResourceType.CLIP_PRESENCE,
    }
    item_cls = Presence

    async def set_config(
        self,
        id: str,
        delay: int | None = None,
        device_mode: PresenceConfigDeviceMode | None = None,
        duration: int | None = None,
        reset_presence: bool | None = None,
        sensitivity: int | None = None,
        trigger_distance: PresenceConfigTriggerDistance | None = None,
    ) -> dict[str, Any]:
        """Change config of presence sensor.

        Supported values:
        - delay [int] 0-65535 (in seconds)
        - device_mode [str]
          - leftright
          - undirected
        - duration [int] 0-65535 (in seconds)
        - reset_presence [bool] True/False
        - sensitivity [int] 0-[sensitivitymax]
        - trigger_distance [str]
          - far
          - medium
          - near
        """
        data: dict[str, int | str] = {
            key: value
            for key, value in {
                "delay": delay,
                "duration": duration,
                "resetpresence": reset_presence,
                "sensitivity": sensitivity,
            }.items()
            if value is not None
        }
        if device_mode is not None:
            data["devicemode"] = device_mode.value
        if trigger_distance is not None:
            data["triggerdistance"] = trigger_distance.value
        return await self.gateway.request_with_retry(
            "put",
            path=f"{self.path}/{id}/config",
            json=data,
        )


class PressureHandler(APIHandler[Pressure]):
    """Handler for pressure sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_types = {
        ResourceType.ZHA_PRESSURE,
        ResourceType.CLIP_PRESSURE,
    }
    item_cls = Pressure


class RelativeRotaryHandler(APIHandler[RelativeRotary]):
    """Handler for relative rotary sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_RELATIVE_ROTARY
    item_cls = RelativeRotary


class SwitchHandler(APIHandler[Switch]):
    """Handler for switch sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_types = {
        ResourceType.ZHA_SWITCH,
        ResourceType.ZGP_SWITCH,
        ResourceType.CLIP_SWITCH,
    }
    item_cls = Switch

    async def set_config(
        self,
        id: str,
        device_mode: SwitchDeviceMode | None = None,
        mode: SwitchMode | None = None,
        window_covering_type: SwitchWindowCoveringType | None = None,
    ) -> dict[str, Any]:
        """Change config of presence sensor.

        Supported values:
        - device_mode [SwitchDeviceMode]
          - "dualpushbutton"
          - "dualrocker"
          - "singlepushbutton"
          - "singlerocker"
        - mode [SwitchMode]
          - "momentary"
          - "rocker"
        - window_covering_type [SwitchWindowCoveringType] 0-9
        """
        data: dict[str, int | str] = {}
        if device_mode is not None:
            data["devicemode"] = device_mode.value
        if mode is not None:
            data["mode"] = mode.value
        if window_covering_type is not None:
            data["windowcoveringtype"] = window_covering_type.value
        return await self.gateway.request_with_retry(
            "put",
            path=f"{self.path}/{id}/config",
            json=data,
        )


class TemperatureHandler(APIHandler[Temperature]):
    """Handler for temperature sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_types = {
        ResourceType.ZHA_TEMPERATURE,
        ResourceType.CLIP_TEMPERATURE,
    }
    item_cls = Temperature


class ThermostatHandler(APIHandler[Thermostat]):
    """Handler for thermostat sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_types = {
        ResourceType.ZHA_THERMOSTAT,
        ResourceType.CLIP_THERMOSTAT,
    }
    item_cls = Thermostat

    async def set_config(
        self,
        id: str,
        cooling_setpoint: int | None = None,
        enable_schedule: bool | None = None,
        external_sensor_temperature: int | None = None,
        external_window_open: bool | None = None,
        fan_mode: ThermostatFanMode | None = None,
        flip_display: bool | None = None,
        heating_setpoint: int | None = None,
        locked: bool | None = None,
        mode: ThermostatMode | None = None,
        mounting_mode: bool | None = None,
        on: bool | None = None,
        preset: ThermostatPreset | None = None,
        schedule: list[str] | None = None,
        set_valve: bool | None = None,
        swing_mode: ThermostatSwingMode | None = None,
        temperature_measurement: ThermostatTemperatureMeasurement | None = None,
        window_open_detection: bool | None = None,
    ) -> dict[str, Any]:
        """Change config of thermostat.

        Supported values:
        - cooling_setpoint [int] 700-3500
        - enable_schedule [bool] True/False
        - external_sensor_temperature [int] -32768-32767
        - external_window_open [bool] True/False
        - fan_mode [ThermostatFanMode]
          - "auto"
          - "high"
          - "low"
          - "medium"
          - "off"
          - "on"
          - "smart"
        - flip_display [bool] True/False
        - heating_setpoint [int] 500-3200
        - locked [bool] True/False
        - mode [ThermostatMode]
          - "auto"
          - "cool"
          - "dry"
          - "emergency heating"
          - "fan only"
          - "heat"
          - "off"
          - "precooling"
          - "sleep"
        - mounting_mode [bool] True/False
        - on [bool] True/False
        - preset [ThermostatPreset]
          - "auto"
          - "boost"
          - "comfort"
          - "complex"
          - "eco"
          - "holiday"
          - "manual"
        - schedule [list]
        - set_valve [bool] True/False
        - swing_mode [ThermostatSwingMode]
          - "fully closed"
          - "fully open"
          - "half open"
          - "quarter open"
          - "three quarters open"
        - temperature_measurement [ThermostatTemperatureMeasurement]
          - "air sensor"
          - "floor protection"
          - "floor sensor"
        - window_open_detection [bool] True/False
        """
        data: dict[str, Any] = {
            key: value
            for key, value in {
                "coolsetpoint": cooling_setpoint,
                "schedule_on": enable_schedule,
                "externalsensortemp": external_sensor_temperature,
                "externalwindowopen": external_window_open,
                "displayflipped": flip_display,
                "heatsetpoint": heating_setpoint,
                "locked": locked,
                "mountingmode": mounting_mode,
                "on": on,
                "schedule": schedule,
                "setvalve": set_valve,
                "windowopen_set": window_open_detection,
            }.items()
            if value is not None
        }
        if fan_mode is not None:
            data["fanmode"] = fan_mode.value
        if mode is not None:
            data["mode"] = mode.value
        if preset is not None:
            data["preset"] = preset.value
        if swing_mode is not None:
            data["swingmode"] = swing_mode.value
        if temperature_measurement is not None:
            data["temperaturemeasurement"] = temperature_measurement.value
        return await self.gateway.request_with_retry(
            "put",
            path=f"{self.path}/{id}/config",
            json=data,
        )


class TimeHandler(APIHandler[Time]):
    """Handler for time sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_TIME
    item_cls = Time


class VibrationHandler(APIHandler[Vibration]):
    """Handler for vibration sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_VIBRATION
    item_cls = Vibration


class WaterHandler(APIHandler[Water]):
    """Handler for water sensor."""

    resource_group = ResourceGroup.SENSOR
    resource_type = ResourceType.ZHA_WATER
    item_cls = Water


SensorResources = Union[
    AirPurifier,
    AirQuality,
    Alarm,
    AncillaryControl,
    Battery,
    CarbonMonoxide,
    Consumption,
    Daylight,
    DoorLock,
    Fire,
    GenericFlag,
    GenericStatus,
    Humidity,
    LightLevel,
    Moisture,
    OpenClose,
    Power,
    Presence,
    Pressure,
    RelativeRotary,
    Switch,
    Temperature,
    Thermostat,
    Time,
    Vibration,
    Water,
]


class SensorResourceManager(GroupedAPIHandler[SensorResources]):
    """Represent deCONZ sensors."""

    resource_group = ResourceGroup.SENSOR

    def __init__(self, gateway: DeconzSession) -> None:
        """Initialize sensor manager."""

        self.air_purifier = AirPurifierHandler(gateway, grouped=True)
        self.air_quality = AirQualityHandler(gateway, grouped=True)
        self.alarm = AlarmHandler(gateway, grouped=True)
        self.ancillary_control = AncillaryControlHandler(gateway, grouped=True)
        self.battery = BatteryHandler(gateway, grouped=True)
        self.carbon_monoxide = CarbonMonoxideHandler(gateway, grouped=True)
        self.consumption = ConsumptionHandler(gateway, grouped=True)
        self.daylight = DaylightHandler(gateway, grouped=True)
        self.door_lock = DoorLockHandler(gateway, grouped=True)
        self.fire = FireHandler(gateway, grouped=True)
        self.generic_flag = GenericFlagHandler(gateway, grouped=True)
        self.generic_status = GenericStatusHandler(gateway, grouped=True)
        self.humidity = HumidityHandler(gateway, grouped=True)
        self.light_level = LightLevelHandler(gateway, grouped=True)
        self.open_close = OpenCloseHandler(gateway, grouped=True)
        self.moisture = MoistureHandler(gateway, grouped=True)
        self.power = PowerHandler(gateway, grouped=True)
        self.presence = PresenceHandler(gateway, grouped=True)
        self.pressure = PressureHandler(gateway, grouped=True)
        self.relative_rotary = RelativeRotaryHandler(gateway, grouped=True)
        self.switch = SwitchHandler(gateway, grouped=True)
        self.temperature = TemperatureHandler(gateway, grouped=True)
        self.thermostat = ThermostatHandler(gateway, grouped=True)
        self.time = TimeHandler(gateway, grouped=True)
        self.vibration = VibrationHandler(gateway, grouped=True)
        self.water = WaterHandler(gateway, grouped=True)

        handlers: list[APIHandler[Any]] = [
            self.air_purifier,
            self.air_quality,
            self.alarm,
            self.ancillary_control,
            self.battery,
            self.carbon_monoxide,
            self.consumption,
            self.daylight,
            self.door_lock,
            self.fire,
            self.generic_flag,
            self.generic_status,
            self.humidity,
            self.light_level,
            self.moisture,
            self.open_close,
            self.power,
            self.presence,
            self.pressure,
            self.relative_rotary,
            self.switch,
            self.temperature,
            self.thermostat,
            self.time,
            self.vibration,
            self.water,
        ]

        super().__init__(gateway, handlers)



"""Read attributes from your deCONZ gateway."""

from __future__ import annotations

import argparse
import asyncio
import logging

import aiohttp
import async_timeout

from pydeconz import errors
from pydeconz.gateway import DeconzSession
from pydeconz.interfaces.api_handlers import CallbackType
from pydeconz.models.event import EventType

LOGGER = logging.getLogger(__name__)


def new_device_callback(event: EventType, id: str) -> None:
    """Signal new device is available."""
    LOGGER.info(f"{event}, {id}")


async def deconz_gateway(
    session: aiohttp.ClientSession,
    host: str,
    port: int,
    api_key: str,
    callback: CallbackType,
) -> DeconzSession | None:
    """Create a gateway object and verify configuration."""
    deconz = DeconzSession(session, host, port, api_key)
    deconz.subscribe(callback)

    try:
        async with async_timeout.timeout(5):
            await deconz.refresh_state()
        return deconz

    except errors.Unauthorized:
        LOGGER.exception("Invalid API key for deCONZ gateway")

    except (asyncio.TimeoutError, errors.RequestError):
        LOGGER.error("Error connecting to deCONZ gateway")

    return None


async def main(host: str, port: int, api_key: str) -> None:
    """CLI method for library."""
    LOGGER.info("Starting deCONZ gateway")

    session = aiohttp.ClientSession()

    gateway = await deconz_gateway(
        session=session,
        host=host,
        port=port,
        api_key=api_key,
        callback=new_device_callback,
    )

    if not gateway:
        LOGGER.error("Couldn't connect to deCONZ gateway")
        await session.close()
        return

    await gateway.refresh_state()
    gateway.start()

    try:
        while True:
            await asyncio.sleep(1)

    except asyncio.CancelledError:
        pass

    finally:
        gateway.close()
        await session.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("host", type=str)
    parser.add_argument("api_key", type=str)
    parser.add_argument("-p", "--port", type=int, default=80)
    parser.add_argument("-D", "--debug", action="store_true")
    args = parser.parse_args()

    loglevel = logging.INFO
    if args.debug:
        loglevel = logging.DEBUG
    logging.basicConfig(format="%(message)s", level=loglevel)

    LOGGER.info(f"{args.host}, {args.port}, {args.api_key}")

    try:
        asyncio.run(
            main(
                host=args.host,
                port=args.port,
                api_key=args.api_key,
            )
        )

    except KeyboardInterrupt:
        pass



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

import enum
from typing import Literal, TypedDict

from . import SensorBase


class SwitchDeviceMode(enum.Enum):
    """Different modes for the Hue wall switch module."""

    SINGLE_ROCKER = "singlerocker"
    SINGLE_PUSH_BUTTON = "singlepushbutton"
    DUAL_ROCKER = "dualrocker"
    DUAL_PUSH_BUTTON = "dualpushbutton"


class SwitchMode(enum.Enum):
    """For Ubisys S1/S2, operation mode of the switch."""

    MOMENTARY = "momentary"
    ROCKER = "rocker"


class SwitchWindowCoveringType(enum.IntEnum):
    """Set the covering type and starts calibration for Ubisys J1."""

    ROLLER_SHADE = 0
    ROLLER_SHADE_TWO_MOTORS = 1
    ROLLER_SHADE_EXTERIOR = 2
    ROLLER_SHADE_TWO_MOTORS_EXTERIOR = 3
    DRAPERY = 4
    AWNING = 5
    SHUTTER = 6
    TILT_BLIND_LIFT_ONLY = 7
    TILT_BLIND_LIFT_AND_TILT = 8
    PROJECTOR_SCREEN = 9


class TypedSwitchConfig(TypedDict):
    """Switch config type definition."""

    devicemode: Literal[
        "dualpushbutton", "dualrocker", "singlepushbutton", "singlerocker"
    ]
    mode: Literal["momentary", "rocker"]
    windowcoveringtype: Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


class TypedSwitchState(TypedDict):
    """Switch state type definition."""

    angle: int
    buttonevent: int
    eventduration: int
    gesture: int
    xy: tuple[float, float]


class TypedSwitch(TypedDict):
    """Switch type definition."""

    config: TypedSwitchConfig
    state: TypedSwitchState


class Switch(SensorBase):
    """Switch sensor."""

    raw: TypedSwitch

    @property
    def button_event(self) -> int | None:
        """Button press."""
        return self.raw["state"].get("buttonevent")

    @property
    def gesture(self) -> int | None:
        """Gesture used for Xiaomi magic cube."""
        return self.raw["state"].get("gesture")

    @property
    def angle(self) -> int | None:
        """Angle representing color on a tint remote color wheel."""
        return self.raw["state"].get("angle")

    @property
    def xy(self) -> tuple[float, float] | None:
        """X/Y color coordinates selected on a tint remote color wheel."""
        return self.raw["state"].get("xy")

    @property
    def event_duration(self) -> int | None:
        """Duration of a push button event for the Hue wall switch module.

        Increased with 8 for each x001, and they are issued pretty much every 800ms.
        """
        return self.raw["state"].get("eventduration")

    @property
    def device_mode(self) -> SwitchDeviceMode | None:
        """Different modes for the Hue wall switch module.

        Behavior as rocker:
          Issues a x000/x002 each time you flip the rocker (to either position).
          The event duration for the x002 is 1 (for 100ms),
          but lastupdated suggests it follows the x000 faster than that.

        Behavior as pushbutton:
          Issues a x000/x002 sequence on press/release.
          Issues a x000/x001/.../x001/x003 on press/hold/release.
          Similar to Hue remotes.
        """
        if "devicemode" in self.raw["config"]:
            return SwitchDeviceMode(self.raw["config"]["devicemode"])
        return None

    @property
    def mode(self) -> SwitchMode | None:
        """For Ubisys S1/S2, operation mode of the switch."""
        if "mode" in self.raw["config"]:
            return SwitchMode(self.raw["config"]["mode"])
        return None

    @property
    def window_covering_type(self) -> SwitchWindowCoveringType | None:
        """Set the covering type and starts calibration for Ubisys J1."""
        if "windowcoveringtype" in self.raw["config"]:
            return SwitchWindowCoveringType(self.raw["config"]["windowcoveringtype"])
        return None



"""Python library to connect deCONZ and Home Assistant to work together."""

from typing import TypedDict

from . import ResourceGroup
from .api import APIItem


class TypedScene(TypedDict):
    """Scene type definition."""

    id: str
    lightcount: int
    transitiontime: int
    name: str


class Scene(APIItem):
    """deCONZ scene representation.

    Dresden Elektroniks documentation of scenes in deCONZ
    http://dresden-elektronik.github.io/deconz-rest-doc/scenes/
    """

    raw: TypedScene
    resource_group = ResourceGroup.SCENE

    _group_resource_id: str = ""
    _group_deconz_id: str = ""

    @property
    def group_id(self) -> str:
        """Group ID representation.

        Scene resource ID is a string combined of group ID and scene ID; "gid_scid".
        """
        if self._group_resource_id == "":
            self._group_resource_id = self.resource_id.split("_")[0]
        return self._group_resource_id

    @property
    def group_deconz_id(self) -> str:
        """Group deCONZ ID representation."""
        if self._group_deconz_id == "":
            self._group_deconz_id = f"/{ResourceGroup.GROUP.value}/{self.group_id}"
        return self._group_deconz_id

    @property
    def deconz_id(self) -> str:
        """Id to call scene over API e.g. /groups/1/scenes/1."""
        return f"{self.group_deconz_id}/{self.resource_group.value}/{self.id}"

    @property
    def id(self) -> str:
        """Scene ID."""
        return self.raw["id"]

    @property
    def light_count(self) -> int:
        """Lights in group."""
        return self.raw["lightcount"]

    @property
    def transition_time(self) -> int:
        """Transition time for scene."""
        return self.raw["transitiontime"]

    @property
    def name(self) -> str:
        """Scene name."""
        return self.raw["name"]



"""Python library to connect deCONZ and Home Assistant to work together."""

from typing import TypedDict

from . import SensorBase


class TypedTimeState(TypedDict):
    """Time state type definition."""

    lastset: str


class TypedTime(TypedDict):
    """Time type definition."""

    state: TypedTimeState


class Time(SensorBase):
    """Time sensor."""

    raw: TypedTime

    @property
    def last_set(self) -> str:
        """Last time time was set."""
        return self.raw["state"]["lastset"]



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

from typing import Any

from ..models import ResourceGroup
from ..models.alarm_system import (
    AlarmSystem,
    AlarmSystemArmAction,
    AlarmSystemDeviceTrigger,
)
from .api_handlers import APIHandler


class AlarmSystems(APIHandler[AlarmSystem]):
    """Manager of deCONZ alarm systems."""

    item_cls = AlarmSystem
    resource_group = ResourceGroup.ALARM

    async def create_alarm_system(self, name: str) -> dict[str, Any]:
        """Create a new alarm system.

        After creation the arm mode is set to disarmed.
        """
        return await self.gateway.request(
            "post",
            path=self.path,
            json={"name": name},
        )

    async def set_alarm_system_configuration(
        self,
        id: str,
        code0: str | None = None,
        armed_away_entry_delay: int | None = None,
        armed_away_exit_delay: int | None = None,
        armed_away_trigger_duration: int | None = None,
        armed_night_entry_delay: int | None = None,
        armed_night_exit_delay: int | None = None,
        armed_night_trigger_duration: int | None = None,
        armed_stay_entry_delay: int | None = None,
        armed_stay_exit_delay: int | None = None,
        armed_stay_trigger_duration: int | None = None,
        disarmed_entry_delay: int | None = None,
        disarmed_exit_delay: int | None = None,
    ) -> dict[str, Any]:
        """Set config of alarm system."""
        data = {
            key: value
            for key, value in {
                "code0": code0,
                "armed_away_entry_delay": armed_away_entry_delay,
                "armed_away_exit_delay": armed_away_exit_delay,
                "armed_away_trigger_duration": armed_away_trigger_duration,
                "armed_night_entry_delay": armed_night_entry_delay,
                "armed_night_exit_delay": armed_night_exit_delay,
                "armed_night_trigger_duration": armed_night_trigger_duration,
                "armed_stay_entry_delay": armed_stay_entry_delay,
                "armed_stay_exit_delay": armed_stay_exit_delay,
                "armed_stay_trigger_duration": armed_stay_trigger_duration,
                "disarmed_entry_delay": disarmed_entry_delay,
                "disarmed_exit_delay": disarmed_exit_delay,
            }.items()
            if value is not None
        }
        return await self.gateway.request(
            "put",
            path=f"{self.path}/{id}/config",
            json=data,
        )

    async def arm(
        self,
        id: str,
        action: AlarmSystemArmAction,
        pin_code: str,
    ) -> dict[str, Any]:
        """Set the alarm to away."""
        return await self.gateway.request(
            "put",
            path=f"{self.path}/{id}/{action.value}",
            json={"code0": pin_code},
        )

    async def add_device(
        self,
        id: str,
        unique_id: str,
        armed_away: bool = False,
        armed_night: bool = False,
        armed_stay: bool = False,
        trigger: AlarmSystemDeviceTrigger | None = None,
        is_keypad: bool = False,
    ) -> dict[str, Any]:
        """Link device with alarm system.

        A device can be linked to exactly one alarm system.
          If it is added to another alarm system, it is automatically removed
          from the prior one.
        This request is used for adding and also for updating a device entry.
        The uniqueid refers to sensors, lights or keypads.
          Adding a light can be useful, e.g. when an alarm should be triggered,
          after a light is powered or switched on in the basement.
        For keypads and keyfobs the request body can be an empty object.
        """
        data = {"armmask": ""}
        data["armmask"] += "A" if armed_away else ""
        data["armmask"] += "N" if armed_night else ""
        data["armmask"] += "S" if armed_stay else ""

        if trigger:
            data["trigger"] = trigger.value

        if is_keypad:
            data = {}

        return await self.gateway.request(
            "put",
            path=f"{self.path}/{id}/device/{unique_id}",
            json=data,
        )

    async def remove_device(self, id: str, unique_id: str) -> dict[str, Any]:
        """Unlink device with alarm system."""
        return await self.gateway.request(
            "delete",
            path=f"{self.path}/{id}/device/{unique_id}",
        )



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

import enum
from typing import TypedDict

from . import LightBase


class TypedCoverState(TypedDict):
    """Cover state type definition."""

    bri: int
    lift: int
    open: bool
    sat: int
    tilt: int


class TypedCover(TypedDict):
    """Cover type definition."""

    state: TypedCoverState


class CoverAction(enum.Enum):
    """Possible cover actions."""

    CLOSE = enum.auto()
    OPEN = enum.auto()
    STOP = enum.auto()


class Cover(LightBase):
    """Cover and Damper class."""

    raw: TypedCover

    @property
    def is_open(self) -> bool:
        """Is cover open."""
        if "open" not in self.raw["state"]:  # Legacy support
            return self.state is False
        return self.raw["state"]["open"]

    @property
    def lift(self) -> int:
        """Amount of closed position.

        Supported values:
          0-100 - 0 is open / 100 is closed
        """
        if "lift" not in self.raw["state"]:  # Legacy support
            return int(self.raw["state"]["bri"] / 2.54)
        return self.raw["state"]["lift"]

    @property
    def tilt(self) -> int | None:
        """Amount of tilt.

        Supported values:
          0-100 - 0 is open / 100 is closed
        """
        if "tilt" in self.raw["state"]:
            return self.raw["state"]["tilt"]
        elif "sat" in self.raw["state"]:  # Legacy support
            return int(self.raw["state"]["sat"] / 2.54)
        return None

    @property
    def supports_tilt(self) -> bool:
        """Supports tilt."""
        return "tilt" in self.raw["state"]



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

from typing import Literal, TypedDict

from . import LightBase
from .light import LightAlert


class TypedSirenState(TypedDict):
    """Siren state type definition."""

    alert: Literal["lselect", "select", "none"]


class TypedSiren(TypedDict):
    """Siren type definition."""

    state: TypedSirenState


class Siren(LightBase):
    """Siren class."""

    raw: TypedSiren

    @property
    def is_on(self) -> bool:
        """If device is sounding."""
        return self.raw["state"]["alert"] == LightAlert.LONG.value



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

from typing import Final, Literal, TypedDict

from . import ResourceGroup
from .deconz_device import DeconzDevice
from .light.light import Light, LightColorMode, LightEffect
from .scene import TypedScene

COLOR_STATE_ATTRIBUTES: Final = {
    "bri",
    "ct",
    "hue",
    "sat",
    "xy",
    "colormode",
    "effect",
}


class TypedGroupAction(TypedDict):
    """Group action type definition."""

    bri: int
    colormode: Literal["ct", "hs", "xy"]
    ct: int
    effect: Literal["colorloop", "none"]
    hue: int
    on: bool
    sat: int
    xy: tuple[float, float]


class TypedGroupState(TypedDict):
    """Group state type definition."""

    all_on: bool
    any_on: bool


class TypedGroup(TypedDict):
    """Group type definition."""

    action: TypedGroupAction
    devicemembership: list[str]
    hidden: bool
    id: str
    lights: list[str]
    lightsequence: list[str]
    multideviceids: list[str]
    name: str
    scenes: list[TypedScene]
    state: TypedGroupState
    type: str


class Group(DeconzDevice):
    """deCONZ light group representation.

    Dresden Elektroniks documentation of light groups in deCONZ
    http://dresden-elektronik.github.io/deconz-rest-doc/groups/
    """

    raw: TypedGroup
    resource_group = ResourceGroup.GROUP

    @property
    def state(self) -> bool | None:
        """Is any light in light group on."""
        return self.any_on

    @property
    def brightness(self) -> int | None:
        """Brightness of the light.

        Depending on the light type 0 might not mean visible "off"
        but minimum brightness.
        """
        return self.raw["action"].get("bri")

    @property
    def color_temp(self) -> int | None:
        """Mired color temperature of the light. (2000K - 6500K)."""
        return self.raw["action"].get("ct")

    @property
    def hue(self) -> int | None:
        """Color hue of the light.

        The hue parameter in the HSV color model is between 0°-360°
        and is mapped to 0..65535 to get 16-bit resolution.
        """
        return self.raw["action"].get("hue")

    @property
    def saturation(self) -> int | None:
        """Color saturation of the light.

        There 0 means no color at all and 255 is the greatest saturation
        of the color.
        """
        return self.raw["action"].get("sat")

    @property
    def xy(self) -> tuple[float, float] | None:
        """CIE xy color space coordinates as array [x, y] of real values (0..1)."""
        x, y = self.raw["action"].get("xy", (None, None))

        if x is None or y is None:
            return None

        if x > 1:
            x = x / 65555

        if y > 1:
            y = y / 65555

        return (x, y)

    @property
    def color_mode(self) -> LightColorMode | None:
        """Color mode of group."""
        if "colormode" in self.raw["action"]:
            return LightColorMode(self.raw["action"]["colormode"])
        return None

    @property
    def effect(self) -> LightEffect | None:
        """Effect of the group."""
        if "effect" in self.raw["action"]:
            return LightEffect(self.raw["action"]["effect"])
        return None

    @property
    def reachable(self) -> bool:
        """Is group reachable."""
        return True

    @property
    def all_on(self) -> bool:
        """Is all lights in light group on."""
        return self.raw["state"].get("all_on") is True

    @property
    def any_on(self) -> bool:
        """Is any lights in light group on."""
        return self.raw["state"].get("any_on") is True

    @property
    def device_membership(self) -> list[str] | None:
        """List of device ids (sensors) when group was created by a device."""
        return self.raw.get("devicemembership")

    @property
    def hidden(self) -> bool | None:
        """Indicate the hidden status of the group.

        Has no effect at the gateway but apps can uses this to hide groups.
        """
        return self.raw.get("hidden")

    @property
    def id(self) -> str | None:
        """Group ID."""
        return self.raw.get("id")

    @property
    def lights(self) -> list[str]:
        """List of all light IDs in group.

        Sequence is defined by the gateway.
        """
        return self.raw.get("lights", [])

    @property
    def light_sequence(self) -> list[str] | None:
        """List of light IDs in group that can be sorted by the user.

        Need not to contain all light ids of this group.
        """
        return self.raw.get("lightsequence")

    @property
    def multi_device_ids(self) -> list[str] | None:
        """List of light IDs in group.

        Subsequent ids from multidevices with multiple endpoints.
        """
        return self.raw.get("multideviceids")

    def update_color_state(
        self, light: Light, update_all_attributes: bool = False
    ) -> None:
        """Sync color state with light.

          update_all_attributes is used to control whether or not to
        write light attributes with the value None to the group.
        This is used to not keep any bad values from the group.
        """
        data = {}

        for attribute in COLOR_STATE_ATTRIBUTES:

            if (light_attribute := light.raw["state"].get(attribute)) is not None:
                data[attribute] = light_attribute
                continue

            if update_all_attributes:
                if attribute == "xy":
                    data[attribute] = (None, None)
                elif attribute == "colormode":
                    continue
                elif attribute == "effect":
                    data[attribute] = LightEffect.NONE.value
                else:
                    data[attribute] = None

        self.update({"action": data})



"""Air purifier data model."""

import enum
from typing import Literal, TypedDict

from . import SensorBase


class TypedAirPurifierState(TypedDict):
    """Air purifier state type definition."""

    deviceruntime: int
    filterruntime: int
    replacefilter: bool
    speed: int


class TypedAirPurifierConfig(TypedDict):
    """Air purifier config type definition."""

    filterlifetime: int
    ledindication: bool
    locked: bool
    mode: Literal[
        "off",
        "auto",
        "speed_1",
        "speed_2",
        "speed_3",
        "speed_4",
        "speed_5",
    ]
    on: bool
    reachable: bool


class TypedAirPurifier(TypedDict):
    """Air purifier type definition."""

    config: TypedAirPurifierConfig
    state: TypedAirPurifierState


class AirPurifierFanMode(enum.Enum):
    """Air purifier supported fan modes."""

    OFF = "off"
    AUTO = "auto"
    SPEED_1 = "speed_1"
    SPEED_2 = "speed_2"
    SPEED_3 = "speed_3"
    SPEED_4 = "speed_4"
    SPEED_5 = "speed_5"


class AirPurifier(SensorBase):
    """Air purifier sensor."""

    raw: TypedAirPurifier

    @property
    def device_run_time(self) -> int:
        """Device run time in minutes."""
        return self.raw["state"]["deviceruntime"]

    @property
    def fan_mode(self) -> AirPurifierFanMode:
        """Fan mode."""
        return AirPurifierFanMode(self.raw["config"]["mode"])

    @property
    def fan_speed(self) -> int:
        """Fan speed."""
        return self.raw["state"]["speed"]

    @property
    def filter_life_time(self) -> int:
        """Filter life time in minutes."""
        return self.raw["config"]["filterlifetime"]

    @property
    def filter_run_time(self) -> int:
        """Filter run time in minutes."""
        return self.raw["state"]["filterruntime"]

    @property
    def led_indication(self) -> bool:
        """Led indicator."""
        return self.raw["config"]["ledindication"]

    @property
    def locked(self) -> bool:
        """Locked configuration."""
        return self.raw["config"]["locked"]

    @property
    def replace_filter(self) -> bool:
        """Replace filter property."""
        return self.raw["state"]["replacefilter"]



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

import enum
import logging
from typing import Literal, TypedDict

from . import LightBase

LOGGER = logging.getLogger(__name__)


class TypedLightState(TypedDict):
    """Light state type definition."""

    alert: Literal[
        "none",
        "select",
        "lselect",
        "blink",
        "breathe",
        "channelchange",
        "finish",
        "okay",
        "stop",
    ]
    bri: int
    colormode: Literal["ct", "hs", "xy"]
    ct: int
    effect: Literal[
        "colorloop",
        "none",
        "carnival",
        "collide",
        "fading",
        "fireworks",
        "flag",
        "glow",
        "rainbow",
        "snake",
        "snow",
        "sparkles",
        "steady",
        "strobe",
        "twinkle",
        "updown",
        "vintage",
        "waves",
    ]
    hue: int
    on: bool
    sat: int
    speed: Literal[0, 1, 2, 3, 4, 5, 6]
    xy: tuple[float, float]


class TypedLight(TypedDict):
    """Light type definition."""

    colorcapabilities: int
    ctmax: int
    ctmin: int
    state: TypedLightState


class LightAlert(enum.Enum):
    """Temporary alert effect.

    Supported values:
    - "none" — light is not performing an alert.
    - "lselect" — light is blinking a longer time.
    - "select" — light is blinking a short time.
    - "blink"
    - "breathe"
    - "channelchange"
    - "finish"
    - "okay"
    - "stop"
    """

    NONE = "none"
    LONG = "lselect"
    SHORT = "select"

    # Specific to Hue color bulbs

    BLINK = "blink"
    BREATHE = "breathe"
    CHANNEL_CHANGE = "channelchange"
    FINISH = "finish"
    OKAY = "okay"
    STOP = "stop"

    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value: object) -> "LightAlert":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unexpected light alert type %s", value)
        return LightAlert.UNKNOWN


class LightColorCapability(enum.IntFlag):
    """Bit field of features supported by a light device."""

    HUE_SATURATION = 0
    ENHANCED_HUE = 1
    COLOR_LOOP = 2
    XY_ATTRIBUTES = 4
    COLOR_TEMPERATURE = 8

    UNKNOWN = 1111

    @classmethod
    def _missing_(cls, value: object) -> "LightColorCapability":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unexpected light color capability %s", value)
        return LightColorCapability.UNKNOWN


class LightColorMode(enum.Enum):
    """Color mode of the light.

    Supported values:
    - "ct" — color temperature.
    - "hs" — hue and saturation.
    - "xy" — CIE xy values.
    """

    CT = "ct"
    HS = "hs"
    XY = "xy"

    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value: object) -> "LightColorMode":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unexpected light color mode %s", value)
        return LightColorMode.UNKNOWN


class LightEffect(enum.Enum):
    """Effect of the light.

    Supported values:
    - "colorloop" — cycle through hue values 0-360
    - "none" — no effect
    - "carnival"
    - "collide"
    - "fading"
    - "fireworks"
    - "flag"
    - "glow"
    - "rainbow"
    - "snake"
    - "snow"
    - "sparkles"
    - "steady"
    - "strobe"
    - "twinkle"
    - "updown"
    - "vintage"
    - "waves"
    """

    COLOR_LOOP = "colorloop"
    NONE = "none"

    # Specific to Lidl christmas light

    CARNIVAL = "carnival"
    COLLIDE = "collide"
    FADING = "fading"
    FIREWORKS = "fireworks"
    FLAG = "flag"
    GLOW = "glow"
    RAINBOW = "rainbow"
    SNAKE = "snake"
    SNOW = "snow"
    SPARKLES = "sparkles"
    STEADY = "steady"
    STROBE = "strobe"
    TWINKLE = "twinkle"
    UPDOWN = "updown"
    VINTAGE = "vintage"
    WAVES = "waves"

    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value: object) -> "LightEffect":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unexpected light effect type %s", value)
        return LightEffect.UNKNOWN


class LightFanSpeed(enum.IntEnum):
    """Possible fan speeds.

    Supported values:
    - 0 - fan is off
    - 1 - 25%
    - 2 - 50%
    - 3 - 75%
    - 4 - 100%
    - 5 - Auto
    - 6 - "comfort-breeze"
    """

    OFF = 0
    PERCENT_25 = 1
    PERCENT_50 = 2
    PERCENT_75 = 3
    PERCENT_100 = 4
    AUTO = 5
    COMFORT_BREEZE = 6


class Light(LightBase):
    """deCONZ light representation.

    Dresden Elektroniks documentation of lights in deCONZ
    http://dresden-elektronik.github.io/deconz-rest-doc/lights/
    """

    raw: TypedLight

    @property
    def alert(self) -> LightAlert | None:
        """Temporary alert effect."""
        if "alert" in self.raw["state"]:
            return LightAlert(self.raw["state"]["alert"])
        return None

    @property
    def brightness(self) -> int | None:
        """Brightness of the light.

        Depending on the light type 0 might not mean visible "off"
        but minimum brightness.
        """
        return self.raw["state"].get("bri")

    @property
    def color_capabilities(self) -> LightColorCapability | None:
        """Bit field to specify color capabilities of light."""
        if "colorcapabilities" in self.raw:
            return LightColorCapability(self.raw["colorcapabilities"])
        return None

    @property
    def color_temp(self) -> int | None:
        """Mired color temperature of the light. (2000K - 6500K)."""
        return self.raw["state"].get("ct")

    @property
    def hue(self) -> int | None:
        """Color hue of the light.

        The hue parameter in the HSV color model is between 0°-360°
        and is mapped to 0..65535 to get 16-bit resolution.
        """
        return self.raw["state"].get("hue")

    @property
    def saturation(self) -> int | None:
        """Color saturation of the light.

        There 0 means no color at all and 255 is the greatest saturation
        of the color.
        """
        return self.raw["state"].get("sat")

    @property
    def xy(self) -> tuple[float, float] | None:
        """CIE xy color space coordinates as array [x, y] of real values (0..1)."""
        x, y = self.raw["state"].get("xy", (None, None))

        if x is None or y is None:
            return None

        if x > 1:
            x = x / 65555

        if y > 1:
            y = y / 65555

        return (x, y)

    @property
    def color_mode(self) -> LightColorMode | None:
        """Color mode of light."""
        if "colormode" in self.raw["state"]:
            return LightColorMode(self.raw["state"]["colormode"])
        return None

    @property
    def on(self) -> bool:
        """Device state."""
        return self.raw["state"]["on"]

    @property
    def max_color_temp(self) -> int | None:
        """Max value for color temperature."""
        if (ctmax := self.raw.get("ctmax")) is not None and ctmax > 650:
            ctmax = 650
        return ctmax

    @property
    def min_color_temp(self) -> int | None:
        """Min value for color temperature."""
        if (ctmin := self.raw.get("ctmin")) is not None and ctmin < 140:
            ctmin = 140
        return ctmin

    @property
    def effect(self) -> LightEffect | None:
        """Effect of the light."""
        if "effect" in self.raw["state"]:
            return LightEffect(self.raw["state"]["effect"])
        return None

    @property
    def fan_speed(self) -> LightFanSpeed:
        """Speed of the fan."""
        return LightFanSpeed(self.raw["state"]["speed"])

    @property
    def supports_fan_speed(self) -> bool:
        """Speed of the fan."""
        return True if "speed" in self.raw["state"] else False



"""Mange events from deCONZ."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Callable, Optional

from ..models import ResourceGroup
from ..models.event import Event, EventType

if TYPE_CHECKING:
    from ..gateway import DeconzSession


LOGGER = logging.getLogger(__name__)

SubscriptionType = tuple[
    Callable[[Event], None],
    Optional[tuple[EventType, ...]],
    Optional[tuple[ResourceGroup, ...]],
]
UnsubscribeType = Callable[[], None]


class EventHandler:
    """Event handler class."""

    def __init__(self, gateway: DeconzSession) -> None:
        """Initialize API items."""
        self.gateway = gateway
        self._subscribers: list[SubscriptionType] = []

    def subscribe(
        self,
        callback: Callable[[Event], None],
        event_filter: tuple[EventType, ...] | EventType | None = None,
        resource_filter: tuple[ResourceGroup, ...] | ResourceGroup | None = None,
    ) -> UnsubscribeType:
        """Subscribe to events.

        "callback" - callback function to call when on event.
        Return function to unsubscribe.
        """
        if isinstance(event_filter, EventType):
            event_filter = (event_filter,)
        if isinstance(resource_filter, ResourceGroup):
            resource_filter = (resource_filter,)

        subscription = (callback, event_filter, resource_filter)
        self._subscribers.append(subscription)

        def unsubscribe() -> None:
            self._subscribers.remove(subscription)

        return unsubscribe

    def handler(self, raw: dict[str, Any]) -> None:
        """Receive event from websocket and pass it along to subscribers."""
        event = Event.from_dict(raw)

        for callback, event_filter, resource_filter in self._subscribers:

            if event_filter is not None and event.type not in event_filter:
                continue

            if resource_filter is not None and event.resource not in resource_filter:
                continue

            callback(event)



"""Python library to connect deCONZ and Home Assistant to work together."""

import enum
from typing import Literal, TypedDict

from . import SensorBase


class DoorLockLockState(enum.Enum):
    """State the lock is in."""

    LOCKED = "locked"
    UNLOCKED = "unlocked"
    UNDEFINED = "undefined"
    NOT_FULLY_LOCKED = "not fully locked"


class TypedDoorLockConfig(TypedDict):
    """Door lock config type definition."""

    lock: bool


class TypedDoorLockState(TypedDict):
    """Door lock state type definition."""

    lockstate: Literal["locked", "unlocked", "undefined", "not fully locked"]


class TypedDoorLock(TypedDict):
    """Door lock type definition."""

    config: TypedDoorLockConfig
    state: TypedDoorLockState


class DoorLock(SensorBase):
    """Door lock sensor."""

    raw: TypedDoorLock

    @property
    def is_locked(self) -> bool:
        """Return True if lock is locked."""
        return self.lock_state == DoorLockLockState.LOCKED

    @property
    def lock_state(self) -> DoorLockLockState:
        """State the lock is in."""
        return DoorLockLockState(self.raw["state"]["lockstate"])

    @property
    def lock_configuration(self) -> bool:
        """Lock configuration."""
        return self.raw["config"]["lock"]



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

import enum
import logging
from typing import Any, Literal, TypedDict

from . import ResourceGroup
from .api import APIItem

LOGGER = logging.getLogger(__name__)


class AlarmSystemArmAction(enum.Enum):
    """Explicit url path to arm and disarm."""

    AWAY = "arm_away"
    NIGHT = "arm_night"
    STAY = "arm_stay"
    DISARM = "disarm"


class AlarmSystemArmMode(enum.Enum):
    """The target arm mode."""

    ARMED_AWAY = "armed_away"
    ARMED_NIGHT = "armed_night"
    ARMED_STAY = "armed_stay"
    DISARMED = "disarmed"


class AlarmSystemArmState(enum.Enum):
    """The current alarm system state."""

    ARMED_AWAY = "armed_away"
    ARMED_NIGHT = "armed_night"
    ARMED_STAY = "armed_stay"
    ARMING_AWAY = "arming_away"
    ARMING_NIGHT = "arming_night"
    ARMING_STAY = "arming_stay"
    DISARMED = "disarmed"
    ENTRY_DELAY = "entry_delay"
    EXIT_DELAY = "exit_delay"
    IN_ALARM = "in_alarm"


class AlarmSystemArmMask(enum.Flag):
    """The target arm mode."""

    ARMED_AWAY = "A"
    ARMED_NIGHT = "N"
    ARMED_STAY = "S"
    NONE = "none"


class AlarmSystemDeviceTrigger(enum.Enum):
    """Specifies arm modes in which the device triggers alarms."""

    ACTION = "state/action"
    BUTTON_EVENT = "state/buttonevent"
    ON = "state/on"
    OPEN = "state/open"
    PRESENCE = "state/presence"
    VIBRATION = "state/vibration"


class TypedAlarmSystemConfig(TypedDict):
    """Alarm system config type definition."""

    armmode: Literal["armed_away", "armed_night", "armed_stay", "disarmed"]
    configured: bool
    disarmed_entry_delay: int
    disarmed_exit_delay: int
    armed_away_entry_delay: int
    armed_away_exit_delay: int
    armed_away_trigger_duration: int
    armed_stay_entry_delay: int
    armed_stay_exit_delay: int
    armed_stay_trigger_duration: int
    armed_night_entry_delay: int
    armed_night_exit_delay: int
    armed_night_trigger_duration: int


class TypedAlarmSystemState(TypedDict):
    """Alarm system state type definition."""

    armstate: Literal[
        "armed_away",
        "armed_night",
        "armed_stay",
        "arming_away",
        "arming_night",
        "arming_stay",
        "disarmed",
        "entry_delay",
        "exit_delay",
        "in_alarm",
    ]
    seconds_remaining: int


class TypedAlarmSystemDevices(TypedDict):
    """Alarm system device type definition."""

    armmask: str
    trigger: str


class TypedAlarmSystem(TypedDict):
    """Alarm system type definition."""

    name: str
    config: TypedAlarmSystemConfig
    state: TypedAlarmSystemState
    devices: dict[str, TypedAlarmSystemDevices]


class AlarmSystem(APIItem):
    """deCONZ alarm system representation.

    Dresden Elektroniks documentation of alarm systems in deCONZ
    https://dresden-elektronik.github.io/deconz-rest-doc/endpoints/alarmsystems/
    """

    raw: TypedAlarmSystem
    resource_group = ResourceGroup.ALARM

    @property
    def arm_state(self) -> AlarmSystemArmState:
        """Alarm system state.

        Can be different from the config.armmode during state transitions.
        """
        return AlarmSystemArmState(self.raw["state"]["armstate"])

    @property
    def seconds_remaining(self) -> int:
        """Remaining time while armstate in "exit_delay" or "entry_delay" state.

        In all other states the value is 0.

        Supported values:
          0-255.
        """
        return self.raw["state"]["seconds_remaining"]

    @property
    def pin_configured(self) -> bool:
        """Is PIN code configured."""
        return self.raw["config"]["configured"]

    @property
    def arm_mode(self) -> AlarmSystemArmMode:
        """Target arm mode."""
        return AlarmSystemArmMode(self.raw["config"]["armmode"])

    @property
    def armed_away_entry_delay(self) -> int:
        """Delay in seconds before an alarm is triggered.

        Supported values:
          0-255.
        """
        return self.raw["config"]["armed_away_entry_delay"]

    @property
    def armed_away_exit_delay(self) -> int:
        """Delay in seconds before an alarm is armed.

        Supported values:
          0-255.
        """
        return self.raw["config"]["armed_away_exit_delay"]

    @property
    def armed_away_trigger_duration(self) -> int:
        """Duration of alarm trigger.

        Supported values:
          0-255.
        """
        return self.raw["config"]["armed_away_trigger_duration"]

    @property
    def armed_night_entry_delay(self) -> int:
        """Delay in seconds before an alarm is triggered.

        Supported values:
          0-255.
        """
        return self.raw["config"]["armed_night_entry_delay"]

    @property
    def armed_night_exit_delay(self) -> int:
        """Delay in seconds before an alarm is armed.

        Supported values:
          0-255.
        """
        return self.raw["config"]["armed_night_exit_delay"]

    @property
    def armed_night_trigger_duration(self) -> int:
        """Duration of alarm trigger.

        Supported values:
          0-255.
        """
        return self.raw["config"]["armed_night_trigger_duration"]

    @property
    def armed_stay_entry_delay(self) -> int:
        """Delay in seconds before an alarm is triggered.

        Supported values:
          0-255.
        """
        return self.raw["config"]["armed_stay_entry_delay"]

    @property
    def armed_stay_exit_delay(self) -> int:
        """Delay in seconds before an alarm is armed.

        Supported values:
          0-255.
        """
        return self.raw["config"]["armed_stay_exit_delay"]

    @property
    def armed_stay_trigger_duration(self) -> int:
        """Duration of alarm trigger.

        Supported values:
          0-255.
        """
        return self.raw["config"]["armed_stay_trigger_duration"]

    @property
    def disarmed_entry_delay(self) -> int:
        """Delay in seconds before an alarm is triggered.

        Supported values:
          0-255.
        """
        return self.raw["config"]["disarmed_entry_delay"]

    @property
    def disarmed_exit_delay(self) -> int:
        """Delay in seconds before an alarm is armed.

        Supported values:
          0-255.
        """
        return self.raw["config"]["disarmed_exit_delay"]

    @property
    def devices(self) -> dict[str, Any]:
        """Devices associated with the alarm system.

        The keys refer to the uniqueid of a light, sensor, or keypad.
        Dictionary values:
        - armmask - A combination of arm modes in which the device triggers alarms.
          A — armed_away
          N — armed_night
          S — armed_stay
          "none" — for keypads and keyfobs
        - trigger - Specifies arm modes in which the device triggers alarms.
          "state/presence"
          "state/open"
          "state/vibration"
          "state/buttonevent"
          "state/on"
        """
        return self.raw["devices"]



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

from asyncio import CancelledError, Task, create_task, sleep
import logging
from pprint import pformat
from typing import Any, Callable

import aiohttp

from .config import Config
from .errors import BridgeBusy, RequestError, ResponseError, raise_error
from .interfaces.alarm_systems import AlarmSystems
from .interfaces.api_handlers import CallbackType, UnsubscribeType
from .interfaces.events import EventHandler
from .interfaces.groups import GroupHandler
from .interfaces.lights import LightResourceManager
from .interfaces.scenes import Scenes
from .interfaces.sensors import SensorResourceManager
from .models import ResourceGroup
from .websocket import Signal, State, WSClient

LOGGER = logging.getLogger(__name__)


class DeconzSession:
    """deCONZ representation that handles lights, groups, scenes and sensors."""

    def __init__(
        self,
        session: aiohttp.ClientSession,
        host: str,
        port: int,
        api_key: str | None = None,
        connection_status: Callable[[bool], None] | None = None,
    ) -> None:
        """Session setup."""
        self.session = session
        self.host = host
        self.port = port
        self.api_key = api_key

        self._sleep_tasks: dict[str, Task[None]] = {}

        self.connection_status_callback = connection_status

        self.config = Config({}, self.request)
        self.events = EventHandler(self)
        self.websocket: WSClient | None = None

        self.alarm_systems = AlarmSystems(self)
        self.groups = GroupHandler(self)
        self.lights = LightResourceManager(self)
        self.scenes = Scenes(self)
        self.sensors = SensorResourceManager(self)

    async def get_api_key(
        self,
        api_key: str | None = None,
        client_name: str = "pydeconz",
    ) -> str:
        """Request a new API key.

        Supported values:
        - api_key [str] 10-40 characters, key to use for authentication
        - client_name [str] 0-40 characters, name of the client application
        """
        data = {
            key: value
            for key, value in {
                "username": api_key,
                "devicetype": client_name,
            }.items()
            if value is not None
        }
        response: list[dict[str, dict[str, str]]] = await self._request(
            "post",
            url=f"http://{self.host}:{self.port}/api",
            json=data,
        )

        return response[0]["success"]["username"]

    def start(self, websocketport: int | None = None) -> None:
        """Connect websocket to deCONZ."""
        if self.config.websocket_port is not None:
            websocketport = self.config.websocket_port

        if not websocketport:
            LOGGER.error("No websocket port specified")
            return

        self.websocket = WSClient(
            self.session, self.host, websocketport, self.session_handler
        )
        self.websocket.start()

    def close(self) -> None:
        """Close websession and websocket to deCONZ."""
        if self.websocket:
            self.websocket.stop()

    async def refresh_state(self) -> None:
        """Read deCONZ parameters."""
        data = await self.request("get", "")

        self.config.raw.update(data[ResourceGroup.CONFIG.value])

        self.alarm_systems.process_raw(data.get(ResourceGroup.ALARM.value, {}))
        self.groups.process_raw(data[ResourceGroup.GROUP.value])
        self.lights.process_raw(data[ResourceGroup.LIGHT.value])
        self.sensors.process_raw(data[ResourceGroup.SENSOR.value])

    def subscribe(self, callback: CallbackType) -> UnsubscribeType:
        """Subscribe to status changes for all resources."""
        subscribers = [
            self.alarm_systems.subscribe(callback),
            self.groups.subscribe(callback),
            self.lights.subscribe(callback),
            self.sensors.subscribe(callback),
        ]

        def unsubscribe() -> None:
            for subscriber in subscribers:
                subscriber()

        return unsubscribe

    async def request_with_retry(
        self,
        method: str,
        path: str,
        json: dict[str, Any] | None = None,
        tries: int = 0,
    ) -> dict[str, Any]:
        """Make a request to the API, retry on BridgeBusy error."""
        if sleep_task := self._sleep_tasks.pop(path, None):
            sleep_task.cancel()

        try:
            return await self.request(method, path, json)

        except BridgeBusy:
            LOGGER.debug("Bridge is busy, schedule retry %s %s", path, str(json))

            if (tries := tries + 1) < 3:
                self._sleep_tasks[path] = sleep_task = create_task(sleep(2 ** (tries)))

                try:
                    await sleep_task
                except CancelledError:
                    return {}

                return await self.request_with_retry(method, path, json, tries)

            self._sleep_tasks.pop(path, None)
            raise BridgeBusy

    async def request(
        self,
        method: str,
        path: str,
        json: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Make a request to the API."""
        response: dict[str, Any] = await self._request(
            method,
            url=f"http://{self.host}:{self.port}/api/{self.api_key}{path}",
            json=json,
        )
        return response

    async def _request(
        self,
        method: str,
        url: str,
        json: dict[str, Any] | None = None,
    ) -> Any:
        """Make a request."""
        LOGGER.debug('Sending "%s" "%s" to "%s"', method, json, url)

        try:
            async with self.session.request(method, url, json=json) as res:

                if res.content_type != "application/json":
                    raise ResponseError(
                        "Invalid content type: {} ({})".format(res.content_type, res)
                    )

                response = await res.json()
                LOGGER.debug("HTTP request response: %s", pformat(response))

                _raise_on_error(response)

                return response

        except aiohttp.client_exceptions.ClientError as err:
            raise RequestError(
                "Error requesting data from {}: {}".format(self.host, err)
            ) from None

    async def session_handler(self, signal: Signal) -> None:
        """Signalling from websocket.

        data - new data available for processing.
        state - network state has changed.
        """
        if not self.websocket:
            return

        if signal == Signal.DATA:
            self.events.handler(self.websocket.data)

        elif signal == Signal.CONNECTION_STATE and self.connection_status_callback:
            self.connection_status_callback(self.websocket.state == State.RUNNING)


def _raise_on_error(data: list[dict[str, Any]] | dict[str, Any]) -> None:
    """Check response for error message."""
    if isinstance(data, list) and data:
        data = data[0]

    if isinstance(data, dict) and "error" in data:
        raise_error(data["error"])



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

import enum
from typing import Literal, TypedDict

from . import SensorBase


class TypedPresenceConfig(TypedDict):
    """Presence config type definition."""

    delay: int
    detectionarea: str
    devicemode: Literal["leftright", "undirected"]
    duration: int
    resetpresence: bool
    sensitivity: int
    sensitivitymax: int
    triggerdistance: Literal["far", "medium", "near"]


class TypedPresenceState(TypedDict):
    """Presence state type definition."""

    dark: bool
    presence: bool
    presenceevent: Literal[
        "enter",
        "leave",
        "enterleft",
        "rightleave",
        "enterright",
        "leftleave",
        "approaching",
        "absenting",
        "8",
        "9",
    ]


class TypedPresence(TypedDict):
    """Presence type definition."""

    config: TypedPresenceConfig
    state: TypedPresenceState


class PresenceConfigDeviceMode(enum.Enum):
    """Device mode.

    Supported values:
    - leftright - left and right
    - undirected
    """

    LEFT_AND_RIGHT = "leftright"
    UNDIRECTED = "undirected"


class PresenceConfigSensitivity(enum.IntEnum):
    """Device sensitivity.

    Supported values:
    - 1 - Low
    - 2 - Medium
    - 3 - High
    """

    LOW = 1
    MEDIUM = 2
    HIGH = 3


class PresenceConfigTriggerDistance(enum.Enum):
    """Trigger distance.

    Supported values:
    - far - Someone approaching is detected on high distance
    - medium - Someone approaching is detected on medium distance
    - near - Someone approaching is detected on low distance
    """

    FAR = "far"
    MEDIUM = "medium"
    NEAR = "near"


class PresenceStatePresenceEvent(enum.Enum):
    """Current activity associated with current presence state.

    Supported values:
    - enter
    - leave
    - enterleft
    - rightleave
    - enterright
    - leftleave
    - approaching
    - absenting
    - 8
    - 9
    """

    ENTER = "enter"
    LEAVE = "leave"
    ENTER_LEFT = "enterleft"
    RIGHT_LEAVE = "rightleave"
    ENTER_RIGHT = "enterright"
    LEFT_LEAVE = "leftleave"
    APPROACHING = "approaching"
    ABSENTING = "absenting"
    EIGHT = "8"
    NINE = "9"


class Presence(SensorBase):
    """Presence detector."""

    raw: TypedPresence

    @property
    def dark(self) -> bool | None:
        """If the area near the sensor is light or not."""
        return self.raw["state"].get("dark")

    @property
    def delay(self) -> int | None:
        """Occupied to unoccupied delay in seconds."""
        return self.raw["config"].get("delay")

    @property
    def device_mode(self) -> PresenceConfigDeviceMode | None:
        """Trigger distance."""
        if "devicemode" in self.raw["config"]:
            return PresenceConfigDeviceMode(self.raw["config"]["devicemode"])
        return None

    @property
    def duration(self) -> int | None:
        """Minimum duration which presence will be true."""
        return self.raw["config"].get("duration")

    @property
    def presence(self) -> bool:
        """Motion detected."""
        return self.raw["state"]["presence"]

    @property
    def presence_event(self) -> PresenceStatePresenceEvent | None:
        """Activity associated with current presence state."""
        if "presenceevent" in self.raw["state"]:
            return PresenceStatePresenceEvent(self.raw["state"]["presenceevent"])
        return None

    @property
    def sensitivity(self) -> int | None:
        """Sensitivity setting for Philips Hue motion sensor.

        Supported values:
        - 0-[sensitivitymax]
        """
        return self.raw["config"].get("sensitivity")

    @property
    def max_sensitivity(self) -> int | None:
        """Maximum sensitivity value."""
        return self.raw["config"].get("sensitivitymax")

    @property
    def trigger_distance(self) -> PresenceConfigTriggerDistance | None:
        """Device specific distance setting."""
        if "triggerdistance" in self.raw["config"]:
            return PresenceConfigTriggerDistance(self.raw["config"]["triggerdistance"])
        return None



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

from typing import TypedDict

from . import LightBase


class TypedLockState(TypedDict):
    """Lock state type definition."""

    on: bool


class TypedLock(TypedDict):
    """Lock type definition."""

    state: TypedLockState


class Lock(LightBase):
    """Lock class."""

    raw: TypedLock

    @property
    def is_locked(self) -> bool:
        """State of lock."""
        return self.raw["state"]["on"]



"""Python library to connect deCONZ and Home Assistant to work together."""

import enum
import logging
from typing import Final, TypedDict

from . import SensorBase

LOGGER = logging.getLogger(__name__)


class DayLightStatus(enum.IntEnum):
    """Day light status."""

    NADIR = 100
    NIGHT_END = 110
    NAUTICAL_DAWN = 120
    DAWN = 130
    SUNRISE_START = 140
    SUNRISE_END = 150
    GOLDEN_HOUR_1 = 160
    SOLAR_NOON = 170
    GOLDEN_HOUR_2 = 180
    SUNSET_START = 190
    SUNSET_END = 200
    DUSK = 210
    NAUTICAL_DUSK = 220
    NIGHT_START = 230

    UNKNOWN = 666

    @classmethod
    def _missing_(cls, value: object) -> "DayLightStatus":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unexpected day light value %s", value)
        return DayLightStatus.UNKNOWN


DAYLIGHT_STATUS: Final = {
    DayLightStatus.NADIR: "nadir",
    DayLightStatus.NIGHT_END: "night_end",
    DayLightStatus.NAUTICAL_DAWN: "nautical_dawn",
    DayLightStatus.DAWN: "dawn",
    DayLightStatus.SUNRISE_START: "sunrise_start",
    DayLightStatus.SUNRISE_END: "sunrise_end",
    DayLightStatus.GOLDEN_HOUR_1: "golden_hour_1",
    DayLightStatus.SOLAR_NOON: "solar_noon",
    DayLightStatus.GOLDEN_HOUR_2: "golden_hour_2",
    DayLightStatus.SUNSET_START: "sunset_start",
    DayLightStatus.SUNSET_END: "sunset_end",
    DayLightStatus.DUSK: "dusk",
    DayLightStatus.NAUTICAL_DUSK: "nautical_dusk",
    DayLightStatus.NIGHT_START: "night_start",
    DayLightStatus.UNKNOWN: "unknown",
}


class TypedDaylightConfig(TypedDict):
    """Daylight config type definition."""

    configured: bool
    sunriseoffset: int
    sunsetoffset: int


class TypedDaylightState(TypedDict):
    """Daylight state type definition."""

    dark: bool
    daylight: bool
    status: int
    sunrise: str
    sunset: str


class TypedDaylight(TypedDict):
    """Daylight type definition."""

    config: TypedDaylightConfig
    state: TypedDaylightState


class Daylight(SensorBase):
    """Daylight sensor built into deCONZ software."""

    raw: TypedDaylight

    @property
    def configured(self) -> bool:
        """Is daylight sensor configured."""
        return self.raw["config"]["configured"]

    @property
    def dark(self) -> bool:
        """Is dark."""
        return self.raw["state"]["dark"]

    @property
    def daylight(self) -> bool:
        """Is daylight."""
        return self.raw["state"]["daylight"]

    @property
    def daylight_status(self) -> DayLightStatus:
        """Return the daylight status string."""
        return DayLightStatus(self.raw["state"]["status"])

    @property
    def status(self) -> str:
        """Return the daylight status string."""
        return DAYLIGHT_STATUS[DayLightStatus(self.raw["state"]["status"])]

    @property
    def sunrise(self) -> str:
        """Sunrise."""
        return self.raw["state"]["sunrise"]

    @property
    def sunrise_offset(self) -> int:
        """Sunrise offset.

        -120 to 120.
        """
        return self.raw["config"]["sunriseoffset"]

    @property
    def sunset(self) -> str:
        """Sunset."""
        return self.raw["state"]["sunset"]

    @property
    def sunset_offset(self) -> int:
        """Sunset offset.

        -120 to 120.
        """
        return self.raw["config"]["sunsetoffset"]



"""Data models."""

from enum import Enum
import logging
from typing import TypeVar

from .api import APIItem

LOGGER = logging.getLogger(__name__)

DataResource = TypeVar("DataResource", bound=APIItem)


class ResourceGroup(Enum):
    """Primary endpoints resources are exposed from."""

    ALARM = "alarmsystems"
    CONFIG = "config"
    GROUP = "groups"
    LIGHT = "lights"
    SCENE = "scenes"
    SENSOR = "sensors"


class ResourceType(Enum):
    """Resource types."""

    # Group resources

    GROUP = "LightGroup"

    # Light resources

    # Configuration tool
    CONFIGURATION_TOOL = "Configuration tool"

    # Cover
    LEVEL_CONTROLLABLE_OUTPUT = "Level controllable output"
    WINDOW_COVERING_CONTROLLER = "Window covering controller"
    WINDOW_COVERING_DEVICE = "Window covering device"

    # Light
    COLOR_DIMMABLE_LIGHT = "Color dimmable light"
    COLOR_LIGHT = "Color light"
    COLOR_TEMPERATURE_LIGHT = "Color temperature light"
    EXTENDED_COLOR_LIGHT = "Extended color light"
    DIMMABLE_LIGHT = "Dimmable light"
    DIMMABLE_PLUGIN_UNIT = "Dimmable plug-in unit"
    DIMMER_SWITCH = "Dimmer switch"
    FAN = "Fan"
    ON_OFF_LIGHT = "On/Off light"
    ON_OFF_OUTPUT = "On/Off output"
    ON_OFF_PLUGIN_UNIT = "On/Off plug-in unit"
    SMART_PLUG = "Smart plug"

    # Lock
    DOOR_LOCK = "Door Lock"

    # Range extender
    RANGE_EXTENDER = "Range extender"

    # Siren
    WARNING_DEVICE = "Warning device"

    # Sensor resources

    # Air purifier
    ZHA_AIR_PURIFIER = "ZHAAirPurifier"

    # Air quality
    ZHA_AIR_QUALITY = "ZHAAirQuality"

    # Alarm
    ZHA_ALARM = "ZHAAlarm"

    # Ancillary control
    ZHA_ANCILLARY_CONTROL = "ZHAAncillaryControl"

    # Battery
    ZHA_BATTERY = "ZHABattery"

    # Carbon monoxide
    ZHA_CARBON_MONOXIDE = "ZHACarbonMonoxide"

    # Consumption
    ZHA_CONSUMPTION = "ZHAConsumption"

    # Daylight
    DAYLIGHT = "Daylight"

    # Door lock
    ZHA_DOOR_LOCK = "ZHADoorLock"

    # Fire
    ZHA_FIRE = "ZHAFire"

    # Generic flag
    CLIP_GENERIC_FLAG = "CLIPGenericFlag"

    # Generic status
    CLIP_GENERIC_STATUS = "CLIPGenericStatus"

    # Humidity
    ZHA_HUMIDITY = "ZHAHumidity"
    CLIP_HUMIDITY = "CLIPHumidity"

    # Light level
    ZHA_LIGHT_LEVEL = "ZHALightLevel"
    CLIP_LIGHT_LEVEL = "CLIPLightLevel"

    # Moisture
    ZHA_MOISTURE = "ZHAMoisture"

    # Open close
    ZHA_OPEN_CLOSE = "ZHAOpenClose"
    CLIP_OPEN_CLOSE = "CLIPOpenClose"

    # Power
    ZHA_POWER = "ZHAPower"

    # Presence
    ZHA_PRESENCE = "ZHAPresence"
    CLIP_PRESENCE = "CLIPPresence"

    # Pressure
    ZHA_PRESSURE = "ZHAPressure"
    CLIP_PRESSURE = "CLIPPressure"

    # Relative rotary
    ZHA_RELATIVE_ROTARY = "ZHARelativeRotary"

    # Switch
    ZHA_SWITCH = "ZHASwitch"
    ZGP_SWITCH = "ZGPSwitch"
    CLIP_SWITCH = "CLIPSwitch"

    # Temperature
    ZHA_TEMPERATURE = "ZHATemperature"
    CLIP_TEMPERATURE = "CLIPTemperature"

    # Thermostat
    ZHA_THERMOSTAT = "ZHAThermostat"
    CLIP_THERMOSTAT = "CLIPThermostat"

    # Time
    ZHA_TIME = "ZHATime"

    # Vibration
    ZHA_VIBRATION = "ZHAVibration"

    # Water
    ZHA_WATER = "ZHAWater"

    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value: object) -> "ResourceType":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unsupported device type %s", value)
        return ResourceType.UNKNOWN



"""API base class."""

from __future__ import annotations

from collections.abc import Callable
import logging
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from . import ResourceGroup

LOGGER = logging.getLogger(__name__)

SubscriptionType = Callable[..., None]
UnsubscribeType = Callable[[], None]


class APIItem:
    """Base class for a deCONZ API item."""

    resource_group: ResourceGroup

    def __init__(self, resource_id: str, raw: Any) -> None:
        """Initialize API item."""
        self.resource_id = resource_id
        self.raw = raw

        self.changed_keys: set[str] = set()

        self._callbacks: list[SubscriptionType] = []
        self._subscribers: list[SubscriptionType] = []

    @property
    def deconz_id(self) -> str:
        """Id to call device over API e.g. /sensors/1."""
        return f"/{self.resource_group.value}/{self.resource_id}"

    def register_callback(self, callback: SubscriptionType) -> None:
        """Register callback for signalling."""
        self._callbacks.append(callback)

    def remove_callback(self, callback: SubscriptionType) -> None:
        """Remove callback previously registered."""
        if callback in self._callbacks:
            self._callbacks.remove(callback)

    def subscribe(self, callback: SubscriptionType) -> UnsubscribeType:
        """Subscribe to events.

        Return function to unsubscribe.
        """
        self._subscribers.append(callback)

        def unsubscribe() -> None:
            """Unsubscribe callback."""
            self._subscribers.remove(callback)

        return unsubscribe

    def update(self, raw: dict[str, dict[str, Any]]) -> None:
        """Update input attr in self.

        Store a set of keys with changed values.
        """
        changed_keys = set()

        for k, v in raw.items():
            changed_keys.add(k)

            if isinstance(self.raw.get(k), dict) and isinstance(v, dict):
                changed_keys.update(set(v.keys()))
                self.raw[k].update(v)

            else:
                self.raw[k] = v

        self.changed_keys = changed_keys

        for callback in self._callbacks + self._subscribers:
            callback()



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

from .. import ResourceGroup
from ..deconz_device import DeconzDevice


class LightBase(DeconzDevice):
    """deCONZ light representation.

    Dresden Elektroniks documentation of lights in deCONZ
    http://dresden-elektronik.github.io/deconz-rest-doc/lights/
    """

    resource_group = ResourceGroup.LIGHT

    @property
    def state(self) -> bool | None:
        """Device state."""
        raw: dict[str, bool] = self.raw["state"]
        return raw.get("on")

    @property
    def reachable(self) -> bool:
        """Is light reachable."""
        raw: dict[str, bool] = self.raw["state"]
        return raw["reachable"]



"""Python library to connect deCONZ and Home Assistant to work together."""

from typing import TypedDict

from . import SensorBase


class TypedAlarmState(TypedDict):
    """Alarm state type definition."""

    alarm: bool


class TypedAlarm(TypedDict):
    """Alarm type definition."""

    state: TypedAlarmState


class Alarm(SensorBase):
    """Alarm sensor."""

    raw: TypedAlarm

    @property
    def alarm(self) -> bool:
        """Alarm."""
        return self.raw["state"]["alarm"]



"""Python library to connect deCONZ and Home Assistant to work together."""

from asyncio import create_task, get_running_loop
from collections import deque
from collections.abc import Callable, Coroutine
import enum
import logging
from typing import Any, Final

import aiohttp
import orjson

LOGGER = logging.getLogger(__name__)


class Signal(enum.Enum):
    """What is the content of the callback."""

    CONNECTION_STATE = "state"
    DATA = "data"


class State(enum.Enum):
    """State of the connection."""

    NONE = ""
    RETRYING = "retrying"
    RUNNING = "running"
    STOPPED = "stopped"


RETRY_TIMER: Final = 15


class WSClient:
    """Websocket transport, session handling, message generation."""

    def __init__(
        self,
        session: aiohttp.ClientSession,
        host: str,
        port: int,
        callback: Callable[[Signal], Coroutine[Any, Any, None]],
    ) -> None:
        """Create resources for websocket communication."""
        self.session = session
        self.host = host
        self.port = port
        self.session_handler_callback = callback

        self.loop = get_running_loop()

        self._data: deque[dict[str, Any]] = deque()
        self._state = self._previous_state = State.NONE

    @property
    def data(self) -> dict[str, Any]:
        """Return data from data queue."""
        try:
            return self._data.popleft()
        except IndexError:
            return {}

    @property
    def state(self) -> State:
        """State of websocket."""
        return self._state

    def set_state(self, value: State) -> None:
        """Set state of websocket and store previous state."""
        self._previous_state = self._state
        self._state = value

    def state_changed(self) -> None:
        """Signal state change."""
        create_task(self.session_handler_callback(Signal.CONNECTION_STATE))

    def start(self) -> None:
        """Start websocket and update its state."""
        create_task(self.running())

    async def running(self) -> None:
        """Start websocket connection."""
        if self._state == State.RUNNING:
            return

        url = f"http://{self.host}:{self.port}"

        try:
            async with self.session.ws_connect(url, heartbeat=60) as ws:
                LOGGER.info("Connected to deCONZ (%s)", self.host)
                self.set_state(State.RUNNING)
                self.state_changed()

                async for msg in ws:

                    if self._state == State.STOPPED:
                        await ws.close()
                        break

                    if msg.type == aiohttp.WSMsgType.TEXT:
                        self._data.append(orjson.loads(msg.data))
                        create_task(self.session_handler_callback(Signal.DATA))
                        LOGGER.debug(msg.data)
                        continue

                    if msg.type == aiohttp.WSMsgType.CLOSED:
                        LOGGER.warning("Connection closed (%s)", self.host)
                        break

                    if msg.type == aiohttp.WSMsgType.ERROR:
                        LOGGER.error("Websocket error (%s)", self.host)
                        break

        except aiohttp.ClientConnectorError:
            if self._state != State.RETRYING:
                LOGGER.error("Websocket is not accessible (%s)", self.host)

        except Exception as err:
            if self._state != State.RETRYING:
                LOGGER.error("Unexpected error (%s) %s", self.host, err)

        if self._state != State.STOPPED:
            self.retry()

    def stop(self) -> None:
        """Close websocket connection."""
        self.set_state(State.STOPPED)
        LOGGER.info("Shutting down connection to deCONZ (%s)", self.host)

    def retry(self) -> None:
        """Retry to connect to deCONZ.

        Do an immediate retry without timer and without signalling state change.
        Signal state change only after first retry fails.
        """
        if self._state == State.RETRYING and self._previous_state == State.RUNNING:
            LOGGER.info(
                "Reconnecting to deCONZ (%s) failed, scheduling retry at an interval of %i seconds",
                self.host,
                RETRY_TIMER,
            )
            self.state_changed()

        self.set_state(State.RETRYING)

        if self._previous_state == State.RUNNING:
            LOGGER.info("Reconnecting to deCONZ (%s)", self.host)
            self.start()
            return

        self.loop.call_later(RETRY_TIMER, self.start)



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

import enum
import logging
from typing import Literal, TypedDict

from . import SensorBase

LOGGER = logging.getLogger(__name__)


class TypedAirQualityState(TypedDict):
    """Air quality state type definition."""

    airquality: Literal[
        "excellent",
        "good",
        "moderate",
        "poor",
        "unhealthy",
        "out of scale",
    ]
    airqualityppb: int
    pm2_5: int


class TypedAirQuality(TypedDict):
    """Air quality type definition."""

    state: TypedAirQualityState


class AirQualityValue(enum.Enum):
    """Air quality.

    Supported values:
    - "excellent"
    - "good"
    - "moderate"
    - "poor"
    - "unhealthy"
    - "out of scale"
    """

    EXCELLENT = "excellent"
    GOOD = "good"
    MODERATE = "moderate"
    POOR = "poor"
    UNHEALTHY = "unhealthy"
    OUT_OF_SCALE = "out of scale"

    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value: object) -> "AirQualityValue":
        """Set default enum member if an unknown value is provided."""
        LOGGER.warning("Unexpected air quality value %s", value)
        return AirQualityValue.UNKNOWN


class AirQuality(SensorBase):
    """Air quality sensor."""

    raw: TypedAirQuality

    @property
    def air_quality(self) -> str:  # AirQualityValue:
        """Air quality."""
        return AirQualityValue(self.raw["state"]["airquality"]).value

    @property
    def air_quality_ppb(self) -> int | None:
        """Air quality PPB TVOC."""
        return self.raw["state"].get("airqualityppb")

    @property
    def pm_2_5(self) -> int | None:
        """Air quality PM2.5 (µg/m³)."""
        return self.raw["state"].get("pm2_5")

    @property
    def supports_air_quality_ppb(self) -> bool:
        """Support Air quality PPB reporting."""
        return "airqualityppb" in self.raw["state"]

    @property
    def supports_pm_2_5(self) -> bool:
        """Support Air quality PM2.5 reporting."""
        return "pm2_5" in self.raw["state"]



"""Python library to connect deCONZ and Home Assistant to work together."""

from __future__ import annotations

from typing import TypedDict

from . import SensorBase


class TypedPowerState(TypedDict):
    """Power state type definition."""

    current: int
    power: int
    voltage: int


class TypedPower(TypedDict):
    """Power type definition."""

    state: TypedPowerState


class Power(SensorBase):
    """Power sensor."""

    raw: TypedPower

    @property
    def current(self) -> int | None:
        """Ampere load of device."""
        return self.raw["state"].get("current")

    @property
    def power(self) -> int:
        """Power load of device."""
        return self.raw["state"]["power"]

    @property
    def voltage(self) -> int | None:
        """Voltage draw of device."""
        return self.raw["state"].get("voltage")
